# -*- coding: utf-8 -*-
#    Copyright (C) 2016  Grace Cox (University of Liverpool)
#
#    Released under the MIT license, a copy of which is located at the root of
#    this project.
"""Module containing functions to parse World Data Centre (WDC) files.

Part of the MagPySV package for geomagnetic data analysis. This module provides
various functions to read, parse and manipulate the contents of World Data
Centre (WDC) formatted files containing geomagnetic data and output data to
comma separated values (CSV) files. Also contains functions to read output of
code used for the COV-OBS magnetic field model series by Gillet et al. (links
below).
"""


import datetime as dt
import glob
import os
import pandas as pd
import numpy as np


def wdc_parsefile(fname):
    """Load a WDC datafile and place the contents into a dataframe.

    Load a datafile of WDC hourly geomagnetic data for a single observatory and
    extract the contents. Parses the current WDC file format, but not the
    previous format containing international quiet (Q) or disturbed (D) day
    designation in place of the century field - only the newer format is
    downloaded from the BGS servers. Detailed file format description
    can be found at http://www.wdc.bgs.ac.uk/catalog/format.html

    Args:
        fname (str): path to a WDC datafile.

    Returns:
        data (pandas.DataFrame):
            dataframe containing hourly geomagnetic data. First column is a
            series of datetime objects (in the format yyyy-mm-dd hh:30:00) and
            subsequent columns are the X, Y and Z components of the magnetic
            field at the specified times.
    """
    # New WDC file format
    cols = [(0, 3), (3, 5), (5, 7), (7, 8), (8, 10), (14, 16),
            (16, 20), (20, 116)]
    col_names = [
        'code', 'yr', 'month', 'component', 'day', 'century',
        'base', 'hourly_values']
    types = {
        'code': str, 'year': int, 'month': int, 'component': str,
        'day': int, 'century': int, 'base': int, 'hourly_values': str}
    data = pd.read_fwf(fname, colspecs=cols, names=col_names,
                       converters=types, header=None)
    data['hourly_values'] = data['hourly_values'].apply(
                                                      separate_hourly_vals)
    data = data.set_index(['code', 'yr', 'month', 'component', 'day',
                           'century', 'base'])['hourly_values'].apply(
                           pd.Series).stack()
    data = data.reset_index()
    data.columns = ['code', 'yr', 'month', 'component', 'day', 'century',
                    'base', 'hour', 'hourly_mean_temp']
    data['hourly_mean_temp'] = data['hourly_mean_temp'].astype(float)

    return data


def separate_hourly_vals(hourstring):
    """Separate individual hourly field means from the string containing all
    24 values in the WDC file. Called by wdc_parsefile.

    Args:
        hourstring (str): string containing the hourly magnetic field means
            parsed from a WDC file for a single day.

    Returns:
        hourly_vals_list (list):
            list containing the hourly field values.
    """
    n = 4
    hourly_vals_list = [hourstring[i:i+n] for i in range(0, len(hourstring),
                        n)]
    return hourly_vals_list


def wdc_datetimes(data):
    """Create datetime objects from the fields extracted from a WDC datafile.

    Args:
        data (pandas.DataFrame): needs columns for century, year (yy format),
            month, day and hour. Called by wdc_parsefile.

    Returns:
        data (pandas.DataFrame):
            the same dataframe with a series of datetime objects (in the format
            yyyy-mm-dd hh:30:00) in the first column.
    """
    # Convert the century/yr columns to a year
    data['year'] = 100 * data['century'] + data['yr']

    # Create datetime objects from the century, year, month and day columns of
    # the WDC format data file. The hourly mean is given at half past the hour
    dates = data.apply(lambda x: dt.datetime.strptime(
        "{0} {1} {2} {3} {4}".format(x['year'], x['month'], x['day'],
                                     x['hour'], 30), "%Y %m %d %H %M"), axis=1)
    data.insert(0, 'date', dates)
    data.drop(['year', 'yr', 'century', 'code', 'day', 'month', 'hour'],
              axis=1, inplace=True)

    return data


def wdc_xyz(data):
    """Convert extracted WDC data to hourly X, Y and Z components in nT.

    Missing values (indicated by 9999 in the datafiles) are replaced with NaNs.

    Args:
        data (pandas.DataFrame): dataframe containing columns for datetime
            objects, magnetic field component (D, I, F, H, X, Y or Z), the
            tabular base and hourly mean.

    Returns:
        data (pandas.DataFrame):
            the same dataframe with datetime objects in the first column and
            columns for X, Y and Z components of magnetic field (in nT).
    """
    # Replace missing values with NaNs
    data.replace(9999, np.nan, inplace=True)
    # Group the data by field component, calculate the hourly means and form
    # a dataframe with separate columns for each field component
    data = data.groupby('component').apply(hourly_mean_conversion)
    data.reset_index(drop=True, inplace=True)
    data.drop(['base', 'hourly_mean_temp'], axis=1, inplace=True)
    data = data.pivot_table(index='date', columns='component',
                            values='hourly_mean')
    data.reset_index(inplace=True)

    # Call helper function to convert D and H components to X and Y
    if 'D' in data.columns and 'H' in data.columns:
        data = angles_to_geographic(data)

    # Make sure that the dataframe contains columns for X, Y and Z components,
    # and create a column of NaN values if a component is missing
    if 'X' not in data.columns:
        data['X'] = np.NaN
    if 'Y' not in data.columns:
        data['Y'] = np.NaN
    if 'Z' not in data.columns:
        data['Z'] = np.NaN

    data = data[['date', 'X', 'Y', 'Z']]

    return data


def hourly_mean_conversion(data):
    """Use the tabular base to calculate hourly means in nT or degrees (D, I).

    Uses the tabular base and hourly value from the WDC file to calculate the
    hourly means of magnetic field components. Value is in nT for H, F, X, Y or
    Z components and in degrees for D or I components. Called by wdc_xyz.

    hourly_mean = tabular_base*100 + wdc_hourly_value (for components in nT)

    hourly_mean = tabular_base + wdc_hourly_value/600 (for D and I components)

    Args:
        data (pandas.DataFrame): dataframe containing columns for datetime
            objects, magnetic field component (D, I, F, H, X, Y or Z), the
            tabular base and hourly mean.

    Returns:
        obs_data (pandas.DataFrame):
            dataframe with datetime objects in the first column and hourly
            means of the field components in either nT or degrees (depending on
            the component).
    """
    obs_data = pd.DataFrame()
    for group in data.groupby('component'):

        if group[0] == 'D' or group[0] == 'I':
            group[1]['hourly_mean'] = group[1]['base'] + \
                (1 / 600.0) * group[1]['hourly_mean_temp']
            obs_data = obs_data.append(group[1], ignore_index=True)
        else:
            group[1]['hourly_mean'] = 100.0 * group[1]['base'] + \
                group[1]['hourly_mean_temp']
            obs_data = obs_data.append(group[1], ignore_index=True)
    return obs_data


def angles_to_geographic(data):
    """Use D and H values to calculate the X and Y field components.

    The declination (D) and horizontal intensity (H) relate to the north (Y)
    and east (X) components as follows:

    X = H*cos(D)

    Y = H*sin(D)

    Args:
        data (pandas.DataFrame): dataframe containing columns for datetime
            objects and hourly means of the magnetic field components (D, I, F,
            H, X, Y or Z).

    Returns:
        data (pandas.DataFrame):
            the same dataframe with datetime objects in the first column and
            hourly means of the field components in either nT or degrees
            (depending on the component).
    """
    data.loc[(~np.isnan(data['D']) & ~np.isnan(data['H'])), 'X'] = data.loc[(
        ~np.isnan(data['D']) & ~np.isnan(data['H'])), 'H'] * np.cos(np.deg2rad(
            data.loc[(~np.isnan(data['D']) & ~np.isnan(data['H'])), 'D']))

    data.loc[(~np.isnan(data['D']) & ~np.isnan(data['H'])), 'Y'] = data.loc[(
        ~np.isnan(data['D']) & ~np.isnan(data['H'])), 'H'] * np.sin(np.deg2rad(
            data.loc[(~np.isnan(data['D']) & ~np.isnan(data['H'])), 'D']))

    return data


def wdc_readfile(fname):
    """Wrapper function to call wdc_parsefile, wdc_datetimes and wdc_xyz.

    Args:
        fname (str): path to a WDC datafile.

    Returns:
        data (pandas.DataFrame):
            dataframe containing the data read from the WDC file. First column
            is a series of datetime objects (in the format yyyy-mm-dd hh:30:00)
            and subsequent columns are the X, Y and Z components of the
            magnetic field at the specified times (hourly means).
    """
    rawdata = wdc_parsefile(fname)
    rawdata = wdc_datetimes(rawdata)
    data = wdc_xyz(rawdata)

    return data


def append_wdc_data(*, obs_name, path=None):
    """Append all WDC data for an observatory into a single dataframe.

    Args:
        obs_name (str): observatory name (as 3-digit IAGA code).
        path (str): path to directory containing WDC datafiles. All files for
            the observatory should be located in the same directory.

    Returns:
        data (pandas.DataFrame):
            dataframe containing all available hourly geomagnetic data at a
            single observatory. First column is a series of datetime objects
            (in the format yyyy-mm-dd hh:30:00) and subsequent columns are the
            X, Y and Z components of the magnetic field at the specified times.
    """
    data = pd.DataFrame()

    data_path = os.path.join(path, obs_name.lower() + '*.wdc')
    # Obtain a list of all files containing the observatory name and ending
    # .wdc in the specified directory
    filenames = glob.glob(data_path)
    # Iterate over the files and append them to previous files
    for file in filenames:
        print(file)
        try:
            frame = wdc_readfile(file)
            data = data.append(frame, ignore_index=True)
        except StopIteration:
            pass

    return data


def covobs_parsefile(*, fname, data_type):
    """Loads MF and SV predictions from the COV-OBS geomagnetic field model.

    Load a datafile containing SV/MF predictions from the COV-OBS magnetic
    field model series by Gillet et al. (2013, Geochem. Geophys. Geosyst.,
    https://doi.org/10.1002/ggge.20041;
    2015, Earth, Planets and Space, https://doi.org/10.1186/s40623-015-0225-z)
    field model.

    Args:
        fname (str): path to a COV-OBS datafile.
        data_type (str): specify whether the file contains magnetic field data
            ('mf') or or secular variation data ('sv')

    Returns:
        model_data (pandas.DataFrame):
            dataframe containing hourly geomagnetic data. First column is a
            series of datetime objects (in the format yyyy-mm-dd) and
            subsequent columns are the X, Y and Z components of the SV/MF at
            the specified times.
    """
    model_data = pd.read_csv(fname, sep=r'\s+', header=None,
                             usecols=[0, 1, 2, 3])
    if data_type is 'mf':
        model_data.columns = ["year_decimal", "X", "Y", "Z"]
    else:
        model_data.columns = ["year_decimal", "dX", "dY", "dZ"]
    return model_data


def covobs_datetimes(data):
    """Create datetime objects from COV-OBS field model output file.

    The format output by the field model is year.decimalmonth e.g. 1960.08 is
    Jan 1960.

    Args:
        data (pandas.DataFrame): needs a column for decimal year (in yyyy.mm
            format).

    Returns:
        data (pandas.DataFrame):
            the same dataframe with the decimal year column replced with a
            series of datetime objects in the format yyyy-mm-dd.
    """
    year_temp = np.floor(data.year_decimal.values.astype(
        'float64')).astype('int')

    months = (12 * (data.year_decimal - year_temp) + 1).round().astype('int')

    data.insert(0, 'year', year_temp)
    data.insert(1, 'month', months)

    date = data.apply(lambda x: dt.datetime.strptime(
        "{0} {1}".format(int(x['year']), int(x['month'])), "%Y %m"),
        axis=1)

    data.insert(0, 'date', date)

    data.drop(['year', 'year_decimal', 'month'], axis=1, inplace=True)

    return data


def covobs_readfile(*, fname, data_type):
    """Wrapper function to call covobs_parsefile and covobs_datetimes.

    The COV-OBS code (publically available) can be used to produce synthetic
    observatory time series for other field models if the appropriate spline
    file is used. The output will be of the same format as COV-OBS output and
    can be read using MagPySV.

    Args:
        fname (str): path to a COV-OBS format datafile.
        data_type (str): specify whether the file contains magnetic field data
            ('mf') or or secular variation data ('sv')
    Returns:
        data (pandas.DataFrame):
            dataframe containing the data read from the file. First column is a
            series of datetime objects (in the format yyyy-mm-dd) and
            subsequent columns are the X, Y and Z components of the SV/MF at
            the specified times.
    """

    rawdata = covobs_parsefile(fname=fname, data_type=data_type)
    data = covobs_datetimes(rawdata)

    return data


def wdc_to_hourly_csv(*, wdc_path=None, write_dir, obs_list,
                      print_obs=True):
    """Convert WDC file to X, Y and Z hourly means and save to CSV file.

    Finds WDC hourly data files for all observatories in a directory path
    (assumes data for all observatories are located inside the same directory).
    The BGS downloading app distributes data inside a single directory
    with the naming convention obsyear.wdc where obs is a three digit
    observatory name in lowercase and year is a four digit year,
    e.g. ngk1990.wdc or clf2013.wdc. This function converts the hourly WDC
    format data to hourly X, Y and Z means, appends all years of data for a
    single observatory into a single dataframe and saves the dataframe to a
    CSV file.

    Args:
        wdc_path (str): path to the directory containing datafiles.
        write_dir (str): directory path to which the output CSV files are
            written.
        obs_list (list): list of observatory names (as 3-digit IAGA codes).
        print_obs (bool): choose whether to print each observatory name as the
            function goes through the directories. Useful for checking progress
            as it can take a while to read the whole WDC dataset. Defaults to
            True.
    """
    # Create the output directory if it does not exist
    if not os.path.exists(write_dir):
        os.makedirs(write_dir)
    # Iterate over each given observatory and call append_wdc_data
    for observatory in obs_list:
        if print_obs is True:
            print(observatory)
        wdc_data = append_wdc_data(
            obs_name=observatory,
            path=wdc_path)
        write_csv_data(data=wdc_data, write_dir=write_dir,
                       obs_name=observatory)


def write_csv_data(*, data, write_dir, obs_name, file_prefix=None,
                   decimal_dates=False, header=True):
    """Write dataframe to a CSV file.

    Args:
        data (pandas.DataFrame): data to be written to file.
        write_dir (str): directory path to which the output CSV file is
            written.
        obs_name (str): name of observatory at which the data were obtained.
        file_prefix (str): optional string to prefix the output CSV filenames
            (useful for specifying parameters used to create the dataset etc).
        decimal_dates (bool): optional argument to specify that dates should be
            written in decimal format rather than datetime objects. Defaults to
            False.
        header (bool): option to include header in file. Defaults to True.
    """

    # Create the output directory if it does not exist
    if not os.path.exists(write_dir):
        os.makedirs(write_dir)

    # Convert datetime objects to decimal dates if required
    if decimal_dates is True:
        data.date = data.date.apply(datetime_to_decimal)
    if file_prefix is not None:
        fpath = os.path.join(write_dir, file_prefix + obs_name + '.csv')
    else:
        fpath = os.path.join(write_dir, obs_name + '.csv')
    data.to_csv(fpath, sep=',', na_rep='NA', header=header, index=False)


def read_csv_data(*, fname, data_type):
    """Read dataframe from a CSV file.

    Args:
        fname (str): path to a CSV datafile.
        data_type (str): specify whether the file contains magnetic field data
            ('mf') or or secular variation data ('sv')

    Returns:
        data (pandas.DataFrame):
            dataframe containing the data read from the CSV file.
    """
    if data_type is 'mf':
        col_names = ['date', 'X', 'Y', 'Z']
    else:
        col_names = ['date', 'dX', 'dY', 'dZ']
    data = pd.read_csv(fname, sep=',', header=0, names=col_names,
                       parse_dates=[0])
    return data


def combine_csv_data(*, start_date, end_date, sampling_rate='MS',
                     obs_list, data_path, model_path, day_of_month=1):
    """Read and combine observatory and model SV data for several locations.

    Calls read_csv_data to read observatory data and field model predictions
    for each observatory in a list. The data and predictions for individual
    observatories are combined into their respective large dataframes. The
    first column contains datetime objects and subsequent columns contain X, Y
    and Z secular variation/field components (in groups of three) for all
    observatories.

    Args:
        start_date (datetime.datetime): the start date of the data analysis.
        end_date (datetime.datetime): the end date of the analysis.
        sampling_rate (str): the sampling rate for the period of interest. The
            default is 'MS', which creates a range of dates between the
            specified values at monthly intervals with the day fixed as the
            first of each month. Use 'M' for the final day of each month. Other
            useful options are 'AS' (a series of dates at annual intervals,
            with the day and month fixed at 01 and January respectively) and
            'A' (as for 'AS' but with the day/month fixed as 31 December.)
        obs_list (list): list of observatory names (as 3-digit IAGA codes).
        data_path (str): path to the CSV files containing observatory data.
        model_path (str): path to the CSV files containing model SV data.
        day_of_month (int): For SV data, first differences of
            monthly means have dates at the start of the month (i.e. MF of
            mid-Feb minus MF of mid-Jan should give SV at Feb 1st. For annual
            differences of monthly means the MF of mid-Jan year 2 minus MF of
            mid-Jan year 1 gives SV at mid-July year 1. The dates of COV-OBS
            output default to the first day of the month (compatible with dates
            of monthly first differences SV data, but not with those of
            annual differences). This option is used to set the day part of the
            dates column if required. Default to 1 (all output dataframes
            will have dates set at the first day of the month.)

    Returns:
        (tuple): tuple containing:

        - obs_data (*pandas.DataFrame*):
            dataframe containing SV data for all observatories in obs_list.
        - model_sv_data (*pandas.DataFrame*):
            dataframe containing SV predictions for all observatories in
            obs_list.
        - model_mf_data (*pandas.DataFrame*):
            dataframe containing magnetic field predictions for all
            observatories in obs_list.
    """
    # Initialise the dataframe with the appropriate date range
    dates = pd.date_range(start_date, end_date, freq=sampling_rate)
    obs_data = pd.DataFrame({'date': dates})
    model_sv_data = pd.DataFrame({'date': dates})
    model_mf_data = pd.DataFrame({'date': dates})

    for observatory in obs_list:

        obs_file = observatory + '.csv'
        model_sv_file = 'sv_' + observatory + '.dat'
        model_mf_file = 'mf_' + observatory + '.dat'
        obs_data_temp = read_csv_data(fname=os.path.join(data_path, obs_file),
                                      data_type='sv')
        model_sv_data_temp = covobs_readfile(fname=os.path.join(model_path,
                                             model_sv_file), data_type='sv')
        model_mf_data_temp = covobs_readfile(fname=os.path.join(model_path,
                                             model_mf_file), data_type='mf')

        model_sv_data_temp['date'] = model_sv_data_temp['date'].apply(
            lambda dt: dt.replace(day=1))

        obs_data_temp.rename(
            columns={'dX': 'dX' + '_' + observatory,
                     'dY': 'dY' + '_' + observatory,
                     'dZ': 'dZ' + '_' + observatory}, inplace=True)
        obs_data_temp['date'] = obs_data_temp['date'].apply(
            lambda dt: dt.replace(day=1))
        model_sv_data_temp.rename(
            columns={'dX': 'dX' + '_' + observatory,
                     'dY': 'dY' + '_' + observatory,
                     'dZ': 'dZ' + '_' + observatory}, inplace=True)
        model_mf_data_temp.rename(
            columns={'X': 'X' + '_' + observatory,
                     'Y': 'Y' + '_' + observatory,
                     'Z': 'Z' + '_' + observatory}, inplace=True)
        # Combine the current observatory data with those of other
        # observatories
        if observatory == obs_list[0]:
            obs_data = pd.merge(
                left=obs_data, right=obs_data_temp,
                how='left', on='date')
            model_sv_data = pd.merge(
                left=model_sv_data, right=model_sv_data_temp,
                how='left', on='date')
            model_mf_data = pd.merge(
                left=model_mf_data, right=model_mf_data_temp,
                how='left', on='date')

        else:
            obs_data = pd.merge(
                left=obs_data, right=obs_data_temp,
                how='left', on='date')
            model_sv_data = pd.merge(
                left=model_sv_data, right=model_sv_data_temp,
                how='left', on='date')
            model_mf_data = pd.merge(
                left=model_mf_data, right=model_mf_data_temp,
                how='left', on='date')
    if day_of_month is not 1:
        model_sv_data['date'] = model_sv_data['date'].apply(
            lambda dt: dt.replace(day=day_of_month))
        model_mf_data['date'] = model_sv_data['date']
        obs_data['date'] = model_sv_data['date']
    return obs_data, model_sv_data, model_mf_data


def datetime_to_decimal(date):
    """Convert a datetime object to a decimal year.

    Args:
        date (datetime.datetime): datetime object representing an observation
            time.

    Returns:
        date (float): the same date represented as a decimal year.
    """
    year_start = dt.datetime(date.year, 1, 1)
    year_end = year_start.replace(year=date.year + 1)
    decimal_year = date.year + (date - year_start) / (year_end - year_start)
    return decimal_year


def ae_parsefile(fname):
    """Load a WDC-like format AE file and place contents into a dataframe.

    Load a file of AE (Auroral Electroject)
    index hourly data in the format distributed by the Kyoto WDC at
    http://wdc.kugi.kyoto-u.ac.jp/dstae/index.html and extract the contents.

    Args:
        fname (str): path to a WDC-like formatted AE file.

    Returns:
        data (pandas.DataFrame):
            dataframe containing hourly AE data. First column is a
            series of datetime objects (in the format yyyy-mm-dd hh:30:00) and
            second column contains theAE values at the specified times.
    """
    # AE WDC file format
    cols = [(0, 2), (3, 5), (5, 7), (8, 10), (14, 16),
            (16, 20), (20, 116)]
    col_names = [
        'code', 'yr', 'month', 'day', 'century',
        'base', 'hourly_values']
    types = {
        'code': str, 'year': int, 'month': int,
        'day': int, 'century': int, 'base': int, 'hourly_values': str}
    data = pd.read_fwf(fname, colspecs=cols, names=col_names,
                       converters=types, header=None)
    data = data.loc[data['code'] == "AE"]
    # Separate the hourly values
    try:
        data['hourly_values'] = data['hourly_values'].apply(
                                                    separate_hourly_vals)
    except ValueError:
        data['hourly_values'] = data['hourly_values'].apply(
            separate_hourly_vals_ae)
    data = data.set_index(['code', 'yr', 'month', 'day',
                           'century', 'base'])['hourly_values'].apply(
                           pd.Series).stack()
    data = data.reset_index()
    data.columns = ['code', 'yr', 'month', 'day', 'century',
                    'base', 'hour', 'hourly_mean_temp']
    data['hourly_mean_temp'] = data['hourly_mean_temp'].astype(float)
    return data


def separate_hourly_vals_ae(hourstring):
    """Separate individual hourly field means from the string containing all
    24 values in the AE file. Called by ae_parsefile.

    Args:
        hourstring (str): string containing the hourly AE means parsed from
            a Kyoto WDC-like file for a single day.

    Returns:
        hourly_vals_list (list):
            list containing the hourly AE values.
    """
    n = 4
    if hourstring[0] is not '-' and hourstring[0] is not ' ':
        hourstring = ' ' + hourstring
    hourly_vals_list = [hourstring[i:i+n] for i in range(0, len(hourstring),
                        n)]
    return hourly_vals_list


def ae_readfile(fname):
    """Wrapper function to call ae_parsefile and wdc_datetimes.

    Args:
        fname (str): path to a AE file in Kyoto WDC-like format. Assumes data
            for all years are contained within this file.

    Returns:
        data (pandas.DataFrame):
            dataframe containing the data read from the WDC file. First column
            is a series of datetime objects (in the format yyyy-mm-dd hh:30:00)
            and second column contains AE values at the specified times
            (hourly means).
    """
    data = ae_parsefile(fname)
    data = wdc_datetimes(data)
    data['hourly_mean'] = 100.0 * data['base'] + \
        data['hourly_mean_temp']
    data.drop(['hourly_mean_temp', 'base'], axis=1, inplace=True)
    return data


def append_ae_data(ae_data_path):
    """Append AE data into a single dataframe containing all years.

    Data downloaded from
    ftp://ftp.ngdc.noaa.gov/STP/GEOMAGNETIC_DATA/INDICES/AURORAL_ELECTROJET/HOURLY/
    come in WDC-like format files with one file per year named aeyyyy.wdc (data
    provided by the WDC at Kyoto. Can be downloaded directly from
    http://wdc.kugi.kyoto-u.ac.jp/dstae/index.html)

    Args:
        ae_data_path (str): path to directory containing WDC-like format AE
            datafiles. All AE files should be located in the same directory.

    Returns:
        data (pandas.DataFrame):
            dataframe containing all available hourly AE data. First column is
            a series of datetime objects (in the format yyyy-mm-dd hh:30:00)
            and second column contains AE values at the specified times.
    """
    data = pd.DataFrame()
    # Obtain a list of all files containing 'ae' and ending in .wdc in the
    # specified directory
    filenames = glob.glob(ae_data_path + 'ae*.txt')
    # Iterate over the files and append them to previous files
    for file in filenames:
        print(file)
        try:
            frame = ae_readfile(file)
            data = data.append(frame, ignore_index=True)
        except StopIteration:
            pass

    return data


def ap_readfile(fname):
    """Load an kp/ap file and place the hourly ap values into a dataframe.

    Load a datafile of 3-hourly ap data and extract the contents. Each of the
    3-hourly values for a given day is repeated three times to give an hourly
    mean for all 24 hours of the day. This function is designed to read files
    downloaded from the GFZ, Potsdam server at
    ftp://ftp.gfz-potsdam.de/pub/home/obs/kp-ap/.

    Args:
        fname (str): path to an ap datafile.

    Returns:
        data (pandas.DataFrame):
            dataframe containing hourly ap data. First column is a
            series of datetime objects (in the format yyyy-mm-dd hh:30:00) and
            second column contains ap values at the specified times.
    """
    col_names = ['full_string']
    types = {'full_string': str}
    # Parse the file
    if fname[-8] == '2':
        cols = [(1, 55)]
        data = pd.read_fwf(fname, colspecs=cols, names=col_names,
                           converters=types, header=None)
        data['month'] = data.full_string.str[1:3]
        data['day'] = data.full_string.str[3:5]
        data['hourly_values'] = data.full_string.str[30:]
    else:
        cols = [(0, 55)]
        data = pd.read_fwf(fname, colspecs=cols, names=col_names,
                           converters=types, header=None)
        data['month'] = data.full_string.str[2:4]
        data['day'] = data.full_string.str[4:6]
        data['hourly_values'] = data.full_string.str[32:]
    data.drop(['full_string'], axis=1, inplace=True)
    data['hourly_values'] = data['hourly_values'].apply(
        separate_three_hourly_vals)
    data = data.set_index(['month', 'day'])['hourly_values'].apply(
                               pd.Series).stack()
    data = data.reset_index()
    data.columns = ['month', 'day', 'hour', 'hourly_mean']
    data['hourly_mean'] = data['hourly_mean'].astype(float)
    data['year'] = int(fname[-8:-4])
    dates = data.apply(lambda x: dt.datetime.strptime(
        "{0} {1} {2} {3} {4}".format(x['year'], x['month'], x['day'],
                                     x['hour'], 30), "%Y %m %d %H %M"), axis=1)
    data.insert(0, 'date', dates)
    data.drop(['year', 'day', 'month', 'hour'],
              axis=1, inplace=True)
    return data


def separate_three_hourly_vals(hourstring):
    """Separate 3-hourly ap means from the string containing all 8 values.

    Separate the 8 individual 3-hourly ap means from the string containing all
    values for the day. Each value is repeated 3 times to give a value for each
    hour. Called by ap_readfile.

    Args:
        hourstring (str): string containing the 3-hourly ap means parsed from
            a Kyoto WDC-like file for a single day.

    Returns:
        hourly_vals_list (list):
            list containing the hourly ap values.
    """
    n = 3
    hourly_vals_list = [hourstring[i:i+n] for i in range(0, len(hourstring),
                        n)]
    hourly_vals_list = np.repeat(hourly_vals_list, n)
    return hourly_vals_list


def append_ap_data(ap_data_path):
    """Append ap data into a single dataframe containing all years.

    Data downloaded from ftp://ftp.gfz-potsdam.de/pub/home/obs/kp-ap/wdc/
    come in WDC-like format files with one file per year named kpyyyy.wdc. This
    function concatenates all years into a single dataframe.

    Args:
        ap_data_path (str): path to directory containing WDC-like format ap
            datafiles. All ap files should be located in the same directory.

    Returns:
        data (pandas.DataFrame):
            dataframe containing all available hourly ap data. First column is
            a series of datetime objects (in the format yyyy-mm-dd hh:30:00)
            and second column contains ap values at the specified times.
    """
    data = pd.DataFrame()
    # Obtain a list of all files containing 'ap' and ending in .wdc in the
    # specified directory
    filenames = glob.glob(ap_data_path + 'kp*.wdc')
    # Iterate over all files and append data
    for file in filenames:
        print(file)
        try:
            frame = ap_readfile(file)
            data = data.append(frame, ignore_index=True)
        except StopIteration:
            pass

    return data



# -*- coding: utf-8 -*-
#    Copyright (C) 2016  Grace Cox (University of Liverpool)
#
#    Released under the MIT license, a copy of which is located at the root of
#    this project.
"""Module containing plotting functions.

Part of the MagPySV package for geomagnetic data analysis. This module provides
various plotting functions.
"""


# import datetime as dt
import matplotlib.pyplot as plt
import numpy as np
import os
import pandas as pd
import scipy as sp
import magpysv.tools as tools

# Setup matplotlib to use latex fonts in figure labels if needed
plt.rcParams['text.latex.preamble'] = [r'\usepackage{amsmath}',
                                       r'\usepackage{sansmath}',
                                       r'\sansmath']


def plot_eigenvalues(*, values, fig_size=(8, 6), font_size=12, label_size=16,
                     save_fig=False, write_path=None):
    """Plot eigenvalues of the covariance matrix of SV residuals.

    Produces a plot of the eigenvalues obtained during the principal component
    analysis (PCA) of SV residuals. The largest eigenvalue represents the
    eigendirection with the largest contribution to the residuals (i.e. the
    "noisy" direction.). The smallest eigenvalue represents the
    eigendirection with the smallest contribution to the residuals (the "clean"
    direction). See Wardinski & Holme (2011, GJI,
    https://doi.org/10.1111/j.1365-246X.2011.04988.x) for further details.

    Args:
        values (array): the eigenvalues obtained from the principal component
            analysis of the SV residuals.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    plt.figure(figsize=fig_size)
    plt.semilogy(values, 'bx-')
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.xlabel(r'$i$', fontsize=label_size)
    plt.ylabel(r'$\lambda_i$', fontsize=label_size)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, 'eigenvalues.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()


def plot_eigenvectors(*, obs_names, eigenvecs, fig_size=(8, 6), font_size=12,
                      label_size=16, save_fig=False, write_path=None):
    """Plot eigenvectors of the covariance matrix of SV residuals.

    Produces a plot of the eigenvectors corresponding to the n largest
    eigenvalues of the covariance matrix obtained during PCA of SV residuals,
    where n is the number of eigenvalues used as a proxy for unmodelled
    external field signal. The n eigenvectors corresponding to the n largest
    eigenvalue represent the directions with the largest contribution
    to the residuals (i.e. the "noisiest" directions). See Wardinski & Holme
    (2011, GJI, https://doi.org/10.1111/j.1365-246X.2011.04988.x)
    for further details.

    Args:
        obs_names (list): list of observatory names given as three digit IAGA
            codes.
        eigenvecs (array): the eigenvalues obtained from the principal
        component analysis of the SV residuals.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    # Loop over directions and plot each eigenvector on a separate subplot
    for direction in range(eigenvecs.shape[1]):
        plt.figure(figsize=fig_size)
        plt.plot(np.abs(eigenvecs[::3, direction]), 'bx',
                 np.abs(eigenvecs[1::3, direction]), 'rx',
                 np.abs(eigenvecs[2::3, direction]), 'cx',
                 markersize=10, mew=3)
        plt.ylim(0, 1)
        plt.grid()
        plt.yticks(fontsize=font_size)
        plt.xticks(range(len(obs_names)), obs_names, fontsize=font_size)
        plt.xticks(rotation=60)
        plt.ylabel(r'$\mathbf{{v}}_{%03d}$' % (direction), fontsize=label_size)
        plt.legend(['x direction', 'y direction', 'z direction'],
                   loc='upper right', frameon=False, fontsize=label_size)
        plt.xlabel('Location', fontsize=label_size)

        if save_fig is True:
            # Create the output directory if it does not exist
            if not os.path.exists(write_path):
                os.makedirs(write_path)
            fpath = os.path.join(write_path, 'eigendirection%03d.pdf' % direction)
            plt.savefig(fpath, bbox_inches='tight')
            plt.close()


def plot_mf(*, dates, mf, model, obs, model_name, fig_size=(8, 6),
            font_size=12, label_size=16, plot_legend=True, save_fig=False,
            write_path=None):
    """Plot the SV and model prediction for a single observatory.

    Produces a plot of the X, Y and Z components of the SV and field
    model prediction for a single observatory.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        mf (time series): X, Y and Z components of magnetic field at a single
            location.
        model (time series): X, Y and Z components of the field predicted by a
            field model for the same location as the data.
        obs (str): observatory name given as three digit IAGA code.
        model_name (str): field model name.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        plot_legend (bool): option to include a legend on the plot. Defaults
            to True.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    plt.figure(figsize=fig_size)
    # X component
    plt.subplot(3, 1, 1)
    plt.title(obs, fontsize=label_size)
    plt.gca().xaxis_date()
    plt.plot(dates, mf.ix[:, 0], 'b', dates, model.ix[:, 0], 'r')
    plt.gcf().autofmt_xdate()
    plt.axis('tight')
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.ylabel(r'$X$ (nT)', fontsize=label_size)
    # Y component
    plt.subplot(3, 1, 2)
    plt.gca().xaxis_date()
    plt.plot(dates, mf.ix[:, 1], 'b', dates, model.ix[:, 1], 'r')
    plt.gcf().autofmt_xdate()
    plt.axis('tight')
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.ylabel(r'$Y$ (nT)', fontsize=label_size)
    # Z component
    plt.subplot(3, 1, 3)
    plt.gca().xaxis_date()
    plt.plot(dates, mf.ix[:, 2], 'b', dates, model.ix[:, 2], 'r')
    plt.gcf().autofmt_xdate()
    plt.axis('tight')
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.xlabel('Year', fontsize=label_size)
    plt.ylabel(r'$Z$ (nT)', fontsize=label_size)

    if plot_legend is True:
        plt.legend([obs, model_name], loc='best', frameon=False)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, obs + '.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()


def plot_sv(*, dates, sv, model, obs, model_name, fig_size=(8, 6),
            font_size=12, label_size=16, plot_legend=False, plot_average=False,
            window_length=12, min_samples=3, save_fig=False, write_path=None):
    """Plot the SV and model prediction for a single observatory.

    Produces a plot of the X, Y and Z components of the SV and field
    model prediction for a single observatory.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        sv (time series): X, Y and Z components of SV at a single location.
        model (time series): X, Y and Z components of the SV predicted by a
            field model for the same location as the data.
        obs (str): observatory name given as three digit IAGA code.
        model_name (str): field model name.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        plot_legend (bool): option to include a legend on the plot. Defaults
            to False.
        plot_average (bool): option to include a running average of the SV time
            series on the plot. Defaults to False.
        window_length (int): number of months over which to take the running
            average if this is plotted. Defaults to 12 months.
        min_samples (int): minimum number of non-NaN values that must be
            present in the window in order for the running average to be
            calculated rather than set to NaN. Defaults to 3 (e.g. for monthly
            first differences this means that at least 3 months of data per
            window are required to calculate the 12-month running average.)
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    if plot_average is True:
        plt.figure(figsize=fig_size)
        # X component
        plt.subplot(3, 1, 1)
        plt.title(obs, fontsize=label_size)
        plt.gca().xaxis_date()
        plt.plot(dates, sv.ix[:, 0], 'b', dates, sv.ix[:, 0].rolling(
            window=window_length, center=True, min_periods=min_samples).mean(),
            'r', dates, model.ix[:, 0], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel(r'$\dot{x}$ (nT/yr)', fontsize=label_size)
        # Y component
        plt.subplot(3, 1, 2)
        plt.gca().xaxis_date()
        plt.plot(dates, sv.ix[:, 1], 'b', dates, sv.ix[:, 1].rolling(
            window=window_length, center=True, min_periods=min_samples).mean(),
            'r', dates, model.ix[:, 1], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel(r'$\dot{y}$ (nT/yr)', fontsize=label_size)
        # Z component
        plt.subplot(3, 1, 3)
        plt.gca().xaxis_date()
        plt.plot(dates, sv.ix[:, 2], 'b', dates, sv.ix[:, 2].rolling(
            window=window_length, center=True, min_periods=min_samples).mean(),
            'r', dates, model.ix[:, 2], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.xlabel('Year', fontsize=label_size)
        plt.ylabel(r'$\dot{z}$ (nT/yr)', fontsize=label_size)

        if plot_legend is True:
            plt.legend([obs, 'Running average', model_name], loc='best',
                       frameon=False)
    else:
        plt.figure(figsize=fig_size)
        # X component
        plt.subplot(3, 1, 1)
        plt.title(obs, fontsize=label_size)
        plt.gca().xaxis_date()
        plt.plot(dates, sv.ix[:, 0], 'b', dates, model.ix[:, 0], 'r')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel(r'$\dot{x}$ (nT/yr)', fontsize=label_size)
        # Y component
        plt.subplot(3, 1, 2)
        plt.gca().xaxis_date()
        plt.plot(dates, sv.ix[:, 1], 'b', dates, model.ix[:, 1], 'r')
        plt.gcf().autofmt_xdate()
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.axis('tight')
        plt.ylabel(r'$\dot{y}$ (nT/yr)', fontsize=label_size)
        # Z component
        plt.subplot(3, 1, 3)
        plt.gca().xaxis_date()
        plt.plot(dates, sv.ix[:, 2], 'b', dates, model.ix[:, 2], 'r')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.xlabel('Year', fontsize=label_size)
        plt.ylabel(r'$\dot{z}$ (nT/yr)', fontsize=label_size)

        if plot_legend is True:
            plt.legend([obs, model_name], loc='best', frameon=False)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, 'clean_' + obs + '.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()


def plot_sv_comparison(*, dates, noisy_sv, denoised_sv, model, obs, model_name,
                       fig_size=(8, 6), font_size=12, label_size=16,
                       plot_legend=False, plot_average=False,
                       window_length=12, min_samples=3, save_fig=False,
                       write_path=None, corrected_residuals, residuals,
                       plot_rms=False):
    """Plot noisy/denoised SV and model prediction for a single observatory.

    Produces a plot of the X, Y and Z components of the noisy SV, the denoised
    SV and field model prediction for a single observatory.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        noisy_sv (time series): X, Y and Z components of uncorrected SV at a
            single location.
        denoised_sv (time series): X, Y and Z components of denoised SV at a
            single location.
        residuals (time series): difference between modelled and observed SV.
        corrected_residuals (time series): difference between modelled and
            denoised observed SV.
        model (time series): X, Y and Z components of the SV predicted by a
            field model for the same location as the data.
        model_name (str): field model name.
        obs (str): observatory name given as three digit IAGA code.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        plot_legend (bool): option to include a legend on the plot. Defaults
            to False.
        plot_average (bool): option to include a running average of the SV time
            series on the plot. Defaults to False.
        window_length (int): number of months over which to take the running
            average if this is plotted. Defaults to 12 months.
        min_samples (int): minimum number of non-NaN values that must be
            present in the window in order for the running average to be
            calculated rather than set to NaN. Defaults to 3 (e.g. for monthly
            first differences this means that at least 3 months of data per
            window are required to calculate the 12-month running average.)
        plot_rms (bool): option to calculate the rms before and after denoising
        and display the values on the figure. Defaults to False.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    Returns:
        (tuple): tuple containing:

        - rms_ratio_x (*float*):
            ratio of rms values for X component residuals before and after
            denoising.
        - rms_ratio_y (*float*):
            ratio of rms values for Y component residuals before and after
            denoising.
        - rms_ratio_z (*float*):
            ratio of rms values for Z component residuals before and after
            denoising.
    """
    plt.figure(figsize=fig_size)
    plt.title(obs, fontsize=label_size)

    if plot_rms is True:
        # Calculate the rms before and after denoising
        rms_x_noisy = np.sqrt(np.nanmean(np.square(residuals.ix[:, 0])))
        rms_x_denoised = np.sqrt(np.nanmean(np.square(
            corrected_residuals.ix[:, 0])))
        rms_y_noisy = np.sqrt(np.nanmean(np.square(residuals.ix[:, 1])))
        rms_y_denoised = np.sqrt(np.nanmean(np.square(
            corrected_residuals.ix[:, 1])))
        rms_z_noisy = np.sqrt(np.nanmean(np.square(residuals.ix[:, 2])))
        rms_z_denoised = np.sqrt(np.nanmean(np.square(
            corrected_residuals.ix[:, 2])))

    if plot_average is True:
        # X component
        plt.subplot(3, 1, 1)
        plt.title(obs, fontsize=label_size)
        plt.gca().xaxis_date()
        plt.plot(dates, noisy_sv.ix[:, 0], 'b', dates, denoised_sv.ix[:, 0],
                 'r', dates, denoised_sv.ix[:, 0].rolling(window=window_length,
                 center=True, min_periods=min_samples).mean(), 'c',
                 dates, model.ix[:, 0], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel(r'$\dot{x}$ (nT/yr)', fontsize=label_size)

        if plot_rms is True:
            plt.annotate('rms = ' + "{:.0f}".format(rms_x_noisy),
                         xy=(0.05, 0.9), xycoords='axes fraction',
                         fontsize=font_size, color='b')
            plt.annotate('rms = ' + "{:.0f}".format(rms_x_denoised),
                         xy=(0.05, 0.8), xycoords='axes fraction',
                         fontsize=font_size, color='r')
        # Y component
        plt.subplot(3, 1, 2)
        plt.gca().xaxis_date()
        plt.plot(dates, noisy_sv.ix[:, 1], 'b', dates, denoised_sv.ix[:, 1],
                 'r', dates, denoised_sv.ix[:, 1].rolling(window=window_length,
                 center=True, min_periods=min_samples).mean(), 'c',
                 dates, model.ix[:, 1], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel(r'$\dot{y}$ (nT/yr)', fontsize=label_size)

        if plot_rms is True:
            plt.annotate('rms = ' + "{:.0f}".format(rms_y_noisy),
                         xy=(0.05, 0.9), xycoords='axes fraction',
                         fontsize=font_size, color='b')
            plt.annotate('rms = ' + "{:.0f}".format(rms_y_denoised),
                         xy=(0.05, 0.8), xycoords='axes fraction',
                         fontsize=font_size, color='r')
        # Z component
        plt.subplot(3, 1, 3)
        plt.gca().xaxis_date()
        plt.plot(dates, noisy_sv.ix[:, 2], 'b', dates, denoised_sv.ix[:, 2],
                 'r', dates, denoised_sv.ix[:, 2].rolling(window=window_length,
                 center=True, min_periods=min_samples).mean(), 'c',
                 dates, model.ix[:, 2], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.xlabel('Year', fontsize=label_size)
        plt.ylabel(r'$\dot{z}$ (nT/yr)', fontsize=label_size)

        if plot_rms is True:
            plt.annotate('rms = ' + "{:.0f}".format(rms_z_noisy),
                         xy=(0.05, 0.9), xycoords='axes fraction',
                         fontsize=font_size, color='b')
            plt.annotate('rms = ' + "{:.0f}".format(rms_z_denoised),
                         xy=(0.05, 0.8), xycoords='axes fraction',
                         fontsize=font_size, color='r')

        if plot_legend is True:
            plt.legend([obs, 'Denoised SV', 'Running average', model_name],
                       loc='best', frameon=False)
    else:
        # X component
        plt.subplot(3, 1, 1)
        plt.title(obs, fontsize=label_size)
        plt.gca().xaxis_date()
        plt.plot(dates, noisy_sv.ix[:, 0], 'b', dates, denoised_sv.ix[:, 0],
                 'r', dates, model.ix[:, 0], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel(r'$\dot{x}$ (nT/yr)', fontsize=label_size)

        if plot_rms is True:
            plt.annotate('rms = ' + "{:.0f}".format(rms_x_noisy),
                         xy=(0.05, 0.9), xycoords='axes fraction',
                         fontsize=font_size, color='b')
            plt.annotate('rms = ' + "{:.0f}".format(rms_x_denoised),
                         xy=(0.05, 0.8), xycoords='axes fraction',
                         fontsize=font_size, color='r')
        # Y component
        plt.subplot(3, 1, 2)
        plt.gca().xaxis_date()
        plt.plot(dates, noisy_sv.ix[:, 1], 'b', dates, denoised_sv.ix[:, 1],
                 'r', dates, model.ix[:, 1], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel(r'$\dot{y}$ (nT/yr)', fontsize=label_size)

        if plot_rms is True:
            plt.annotate('rms = ' + "{:.0f}".format(rms_y_noisy),
                         xy=(0.05, 0.9), xycoords='axes fraction',
                         fontsize=font_size, color='b')
            plt.annotate('rms = ' + "{:.0f}".format(rms_y_denoised),
                         xy=(0.05, 0.8), xycoords='axes fraction',
                         fontsize=font_size, color='r')
        # Z component
        plt.subplot(3, 1, 3)
        plt.gca().xaxis_date()
        plt.plot(dates, noisy_sv.ix[:, 2], 'b', dates, denoised_sv.ix[:, 2],
                 'r', dates, model.ix[:, 2], 'k')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.xlabel('Year', fontsize=label_size)
        plt.ylabel(r'$\dot{z}$ (nT/yr)', fontsize=label_size)

        if plot_rms is True:
            plt.annotate('rms = ' + "{:.0f}".format(rms_z_noisy),
                         xy=(0.05, 0.9), xycoords='axes fraction',
                         fontsize=font_size, color='b')
            plt.annotate('rms = ' + "{:.0f}".format(rms_z_denoised),
                         xy=(0.05, 0.8), xycoords='axes fraction',
                         fontsize=font_size, color='r')

        if plot_legend is True:
            plt.legend([obs, 'Denoised SV', model_name],
                       loc='best', frameon=False)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, obs + '.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()

    if plot_rms is True:
        rms_ratio_x = rms_x_denoised/rms_x_noisy
        rms_ratio_y = rms_y_denoised/rms_y_noisy
        rms_ratio_z = rms_z_denoised/rms_z_noisy
        return rms_ratio_x, rms_ratio_y, rms_ratio_z
    else:
        return


def plot_index(*, index_file, dates, projected_residuals, fig_size=(8, 6),
               font_size=12, label_size=16, plot_legend=True, save_fig=False,
               write_path=None, index_name='Dst'):
    """Compare the proxy used to denoise the SV data with a geomagnetic index.

    Loads geomagnetic index and plots it alongside the signal
    used as a proxy for unmodelled external signal. Both time series are
    reduced to zero mean and unit variance (i.e. their zscore) for plotting.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        index_file (str): path to the file containing index data.
        projected_residuals (time series): difference between modelled and
            SV rotated into the eigendirections obtained during denoising
            (principal component analysis). The proxy for unmodelled external
            signal is the residual projected in the noisiest eigendirection(s).
        index_name (str): name of index used in comparison e.g. Dst or ap.
            Defaults to Dst.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        plot_legend (bool): option to include a legend on the plot. Defaults
            to True.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    for direction in range(projected_residuals.shape[1]):
        signal = projected_residuals[:, direction]
        coeff, df = tools.calculate_correlation_index(
            dates=dates, signal=signal, index_file=index_file)
        # Plot the zscore of the two time series
        plt.figure(figsize=fig_size)
        plt.gca().xaxis_date()
        plt.plot(df.date, sp.stats.mstats.zscore(df.index_vals), 'b',
                 dates, sp.stats.mstats.zscore(signal), 'r')
        plt.gcf().autofmt_xdate()
        plt.axis('tight')
        plt.xlabel('Year', fontsize=label_size)
        plt.ylabel('Signal (nT/yr)', fontsize=label_size)
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.annotate('|r| = ' + "{:.2f}".format(np.abs(coeff)), xy=(0.05, 0.9),
                     xycoords='axes fraction', fontsize=16)

        if plot_legend is True:
            plt.legend([index_name, 'proxy'], loc='best', frameon=False,
                       fontsize=16)

        if save_fig is True:
            # Create the output directory if it does not exist
            if not os.path.exists(write_path):
                os.makedirs(write_path)
            fpath = os.path.join(write_path, index_name \
                + '_eigendirection%03d.pdf' % direction)
            plt.savefig(fpath, bbox_inches='tight')
            plt.close()


def plot_index_dft(*, index_file, dates, signal, fig_size=(8, 6), font_size=12,
                   label_size=16, plot_legend=True, save_fig=False,
                   write_path=None, index_name='Dst'):
    """Compare the DFTs of the proxy signal with that of a geomagnetic index.

    Loads index data, calculates its DFT using an FFT algorithm and plots it
    alongside the DFT of the signal used as a proxy for unmodelled external
    signal. The length of the time series are padded with zeroes up to the next
    power of two.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        signal (time series): proxy for unmodelled external signal used in the
            denoising process (principal component analysis). The proxy is the
            residual in the noisiest eigendirection(s).
        index_file (str): path to the file containing index data.
        index_name (str): name of index used in comparison e.g. Dst or Dcx.
            Defaults to Dst.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        plot_legend (bool): option to include a legend on the plot. Defaults
            to True.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    coeff, df = tools.calculate_correlation_index(
        dates=dates, signal=signal, index_file=index_file)
    sampling_period = 1 / 12.0   # Sampling time in years

    # Find the next power of two higher than the length of the time series and
    # perform the DFT with the series padded with zeroes to this length
    sample_length = int(pow(2, np.ceil(np.log2(len(df.proxy)))))
    index_dft = sp.fft(df.index_vals, sample_length)
    proxy_dft = sp.fft(df.proxy, sample_length)
    freq = np.linspace(0.0, 1.0 / (2.0 * sampling_period), sample_length / 2)
    index_power = (2.0 / sample_length) * np.abs(
        index_dft[:sample_length // 2])
    proxy_power = (2.0 / sample_length) * np.abs(
        proxy_dft[:sample_length // 2])

    plt.figure(figsize=fig_size)
    # Time domain
    plt.subplot(2, 1, 1)
    plt.gca().xaxis_date()
    plt.plot(df.date, df.index_vals, 'b', dates, signal, 'r')
    plt.gcf().autofmt_xdate()
    plt.axis('tight')
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.ylabel('Signal (nT/yr)', fontsize=label_size)
    plt.annotate('|r| = ' + "{:.2f}".format(np.abs(coeff)), xy=(0.05, 0.9),
                 xycoords='axes fraction')
    # Frequency domain
    plt.subplot(2, 1, 2)
    plt.plot(freq, index_power, 'b', freq, proxy_power, 'r')
    plt.axis('tight')
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.xlabel('Frequency (cycles/year)', fontsize=label_size)
    plt.ylabel('Power', fontsize=label_size)

    if plot_legend is True:
        plt.legend([index_name, 'proxy'], loc='best', frameon=False)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, index_name + '_dft.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()


def plot_outliers(*, dates, signal, obs_name, outliers, signal_type='SV',
                  fig_size=(8, 6), font_size=12, label_size=16, save_fig=False,
                  write_path=None):
    """Plot the SV and identified outliers.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        signal (time series): single component of SV at a single location.
        obs_name (str): states the SV component and observatory name given as
            three digit IAGA code. For example, the X component at NGK would be
            x_ngk if obs_name is taken from the pandas.DataFrame containing
            SV data for all observatories combined.
        outliers (array): outliers identified by the denoise.detect_outliers
            function
        signal_type (str): specify whether magnetic field ('MF') or secular
            variation ('SV') is plotted. Defaults to SV.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    plt.figure(figsize=fig_size)
    plt.plot(dates, signal, 'k', dates, outliers, 'r^')
    plt.axis('tight')
    plt.xlabel('Year', fontsize=label_size)
    if signal_type is 'SV':
        plt.ylabel('SV (nT/yr)', fontsize=label_size)
    else:
        plt.ylabel('MF (nT)', fontsize=label_size)
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.legend([obs_name, 'outlier'], loc='best', frameon=False)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, obs_name + '_outliers.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()


def plot_residuals_dft(*, projected_residuals, dates, fig_size=(10, 8),
                       font_size=12, label_size=16, plot_legend=True,
                       save_fig=False, write_path=None):
    """Compare the DFTs of the projected residuals with each other.

    Calculates the DFT of the residuals in each eigendirection given and plots
    it alongside the residuals themselves. Produces a single figure with each
    eigendirection included as a subplot. Use plot_residuals_dft_all if a
    separate figure per eigendirection is desired. The length of the time
    series are padded with zeroes up to the next power of two.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        projected_residuals (time series): difference between modelled and
            SV rotated into the eigendirections obtained during denoising
            (principal component analysis). The proxy for unmodelled external
            signal is the residual projected in the noisiest eigendirection(s).
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        plot_legend (bool): option to include a legend on the plot. Defaults to
            True.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    fig_count = 1
    # Create figure with shared subplot axes
    fig, ax = plt.subplots(nrows=projected_residuals.shape[1], ncols=2,
                           sharex=True, sharey=True, figsize=fig_size)

    sampling_period = 1 / 12.0   # Sampling time in years
    sample_length = int(pow(2, np.ceil(np.log2(projected_residuals.shape[0]))))

    # Iterate over the eigendirections and produce a figure for each
    for direction in range(projected_residuals.shape[1]):
        residual_dft = sp.fft(projected_residuals[:, direction], sample_length)
        freq = np.linspace(0.0, 1.0 / (2.0 * sampling_period),
                           sample_length / 2)
        residual_power = (2.0 / sample_length) * np.abs(
            residual_dft[:sample_length // 2])
        plt.subplot(projected_residuals.shape[1], 2, direction + fig_count)
        plt.gca().xaxis_date()
        plt.plot(dates, projected_residuals[:, direction], 'b')
        plt.xticks(rotation=60)
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel('Direction {}'.format(direction), fontsize=label_size-2)
        fig_count = fig_count + 1
        plt.subplot(projected_residuals.shape[1], 2, direction + fig_count)
        # Frequency domain
        plt.plot(freq, residual_power, 'b')
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
    fig.text(0.00, 0.5, 'Residuals (nT/yr)', va='center', rotation='vertical',
             fontsize=label_size)
    fig.text(0.25, 0.02, 'Date', ha='center', fontsize=label_size)
    fig.text(0.75, 0.02, 'Frequency (cycles/yr)', ha='center',
             fontsize=label_size)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, 'residuals_dft.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()


def plot_residuals_dft_all(*, projected_residuals, dates, fig_size=(10, 8),
                           font_size=12, label_size=16,
                           save_fig=False, write_path=None):
    """Compare the DFTs of the projected residuals with each other.

    Calculates the DFT of the residuals in each eigendirection given and plots
    it alongside the residuals themselves. Produces a separate figure per
    eigendirection. Use plot_residuals_dft if a single figure with each
    eigendirection included as a subplot is desired. The length of the time
    series are padded with zeroes up to the next power of two.

    Args:
        dates (datetime.datetime): dates of time series measurements.
        projected_residuals (time series): difference between modelled and
            SV rotated into the eigendirections obtained during denoising
            (principal component analysis). The proxy for unmodelled external
            signal is the residual projected in the noisiest eigendirection(s).
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        plot_legend (bool): option to include a legend on the plot. Defaults to
            True.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """

    sampling_period = 1 / 12.0   # Sampling time in years
    sample_length = int(pow(2, np.ceil(np.log2(projected_residuals.shape[0]))))

    # Iterate over the eigendirections and produce a figure for each
    for direction in range(projected_residuals.shape[1]):
        residual_dft = sp.fft(projected_residuals[:, direction], sample_length)
        freq = np.linspace(0.0, 1.0 / (2.0 * sampling_period),
                           sample_length / 2)
        residual_power = (2.0 / sample_length) * np.abs(
            residual_dft[:sample_length // 2])
        ax = plt.subplots(nrows=1, ncols=2, figsize=fig_size)[1]
        plt.subplot(2, 1, 1)
        plt.gca().xaxis_date()
        plt.plot(dates, projected_residuals[:, direction], 'b')
        plt.xticks(rotation=60)
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.ylabel('Residuals (nT/yr)', fontsize=label_size)
        plt.subplot(2, 1, 2)
        # Frequency domain
        plt.plot(freq, residual_power, 'b')
        plt.axis('tight')
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.xlabel('Frequency (cycles/yr)', fontsize=label_size)
        plt.ylabel('DFT', fontsize=label_size)

        if save_fig is True:
            # Create the output directory if it does not exist
            if not os.path.exists(write_path):
                os.makedirs(write_path)
            fpath = os.path.join(write_path, \
                'dft_eigendirection%03d.pdf' % direction)
            plt.savefig(fpath, bbox_inches='tight')
            plt.close()


def compare_proxies(*, fname1, fname2, legend_text, fig_size=(8, 6),
                    font_size=12, label_size=16,
                    save_fig=False, write_path=None):
    """Compare proxies of unmodelled external signal for different analyses.

    Calculates the correlation coefficients of two given proxies for unmodelled
    external signals and includes it on a plot of the two series. Each
    proxy is formed of the SV residuals projected into the eigendirection(s) of
    the largest eigenvalues of the residual covariance matrix. The proxies are
    reduced to zero-mean and unit-variance on the plots (zscore).

    Args:
        fname1 (str): path to file containing a time series of proxy for noise.
        fname2 (str): path to a second file containing a proxy for noise.
        legend_text (str): text to include on the plot legend.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    proxy1 = pd.read_csv(fname1, parse_dates=[0], names=['date', 'proxy1'],
                         skiprows=1, index_col=None)
    proxy2 = pd.read_csv(fname2, parse_dates=[0], names=['date', 'proxy2'],
                         skiprows=1, index_col=None)
    # Merge the two dataframes so that only dates contained within both are
    # retained
    df = pd.merge(proxy1.dropna(), proxy2.dropna(), on='date', how='inner')
    coeff = np.corrcoef(df.proxy1, df.proxy2)

    # Plot the zscore of the two time series
    plt.figure(figsize=fig_size)
    plt.gca().xaxis_date()
    plt.plot(df.date, sp.stats.mstats.zscore(df.proxy1), 'b',
             df.date, sp.stats.mstats.zscore(df.proxy2), 'r')
    plt.gcf().autofmt_xdate()
    plt.axis('tight')
    plt.xlabel('Year', fontsize=label_size)
    plt.ylabel('Proxy signal (nT/yr)', fontsize=label_size)
    plt.xticks(fontsize=font_size)
    plt.yticks(fontsize=font_size)
    plt.legend(legend_text, loc='upper right', frameon=False, fontsize=16)
    plt.annotate('|r| = ' + "{:.2f}".format(np.abs(coeff.data[0, 1])),
                 xy=(0.05, 0.9),
                 xycoords='axes fraction', fontsize=16)

    if save_fig is True:
        # Create the output directory if it does not exist
        if not os.path.exists(write_path):
            os.makedirs(write_path)
        fpath = os.path.join(write_path, 'proxy_comparison_' + legend_text[0] + '_' \
            + legend_text[1] + '.pdf')
        plt.savefig(fpath, bbox_inches='tight')
        plt.close()


def rms_ratios(*, rms, fig_size=(8, 6), font_size=12, label_size=16,
               save_fig=False, write_path=None):
    """Plot the rms of residuals after removing successive eigendirections.

    Plots the ratio of the residuals rms values before and after denoising, for
    different numbers of eigendirections are removed from the data. Removing
    all eigendirections gives a ratio of zero as the denoised SV now equals the
    model prediction. Requires the dnoising to be run several times, each time
    using a different number of eigendirections for the external signal proxy
    (i.e. different values of the argument proxy_number in calls to
    eigenvalue_analysis.) Uses output from plot_sv_comparison when that
    function is run with the option plot_rms=True.

    Args:
        rms (dict): rms ratios for each component after running
            eigenvalue_analysis with different values for the proxy_number
            argument.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.
        save_fig (bool): option to save figure. Defaults to False.
        write_path (str): output path for figure if saved.
    """
    for observatory in rms.keys():
        plt.figure(figsize=fig_size)
        plt.plot(range(0, 3*len(rms.keys())), rms[observatory]['x'], 'bx-',
                 markersize=7, mew=2)
        plt.plot(range(0, 3*len(rms.keys())), rms[observatory]['y'], 'rx-',
                 markersize=7, mew=2)
        plt.plot(range(0, 3*len(rms.keys())), rms[observatory]['z'], 'cx-',
                 markersize=7, mew=2)
        plt.legend(['x', 'y', 'z'],
                   loc='best', frameon=False, fontsize=16)
        plt.xlabel('Eigendirections removed', fontsize=label_size)
        plt.ylabel('%s rms/denoised rms' % observatory, fontsize=label_size)
        plt.ylim([0, 1.2])
        plt.xlim([0, 3*len(rms.keys())-1])
        plt.xticks(fontsize=font_size)
        plt.yticks(fontsize=font_size)
        plt.grid()

        if save_fig is True:
            # Create the output directory if it does not exist
            if not os.path.exists(write_path):
                os.makedirs(write_path)
            fpath = os.path.join(write_path, 'rms_ratio_%s.pdf' % observatory)
            plt.savefig(fpath, bbox_inches='tight')
            plt.close()



# -*- coding: utf-8 -*-
"""
Created on Sun Feb 21 13:26:22 2016

Testing the file io functionality of io.py.

@author: Grace Cox and Will Brown
"""

import unittest
import mock
from ddt import ddt, data, unpack
from io import StringIO  # io
import os
from .. import io  # magpysv.io
from pandas.util.testing import assert_frame_equal
import pandas as pd
import datetime as dt
import numpy as np

# Directory where the test files are located
TEST_DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')


@ddt
class WDCParsefileTestCase(unittest.TestCase):
    """Set up for test case for parsing function (opens test datafiles)"""
    @data({'filename': 'testdata1.wdc', 'code': 'NGK', 'component1': 'D',
           'component2': 'Z'},
          {'filename': 'testdata2.wdc', 'code': 'NGK', 'component1': 'D',
          'component2': 'Z'},
          {'filename': 'testdata3.wdc', 'code': 'PSM', 'component1': 'H',
          'component2': 'D'})
    @unpack
    def test_wdc_parsefile_newformat(self, filename, code, component1,
                                     component2):
        """"Verify correct parsing of the test WDC format files"""
        testfile = os.path.join(TEST_DATA_PATH, filename)

        data = io.wdc_parsefile(testfile)
        # Observatory code
        self.assertEqual(data.code[0], code)
        self.assertEqual(len(data.code.unique()), 1)
        # Components
        self.assertTrue(any(
            x in {'X', 'Y', 'Z', 'D', 'I', 'H'} for x in data.component))
        self.assertEqual(data.component[3], component1)
        self.assertEqual(data.component.values[-1], component2)


class WDCDatetimesTestCase(unittest.TestCase):
    """Set up the test case for datetimes function"""
    def setUp(self):
        """Create a dataframe containing mock input for datetime function"""
        self.data = pd.DataFrame(
            index=[0], columns=['century', 'yr', 'month', 'day', 'hour'])
        self.data['century'] = 19
        self.data['yr'] = 88
        self.data['month'] = 9
        self.data['day'] = 21
        self.data['hour'] = 2
        self.data['code'] = 'ESK'

    def test_wdc_datetimes(self):
        """Verify correct datetime creation from mock data"""
        df = io.wdc_datetimes(self.data)

        self.assertTrue(isinstance(df.date[0], pd.datetime))
        self.assertEqual(df.date[0], dt.datetime(day=21, month=9, year=1988,
                                                 hour=2, minute=30))


class HourlyMeanConversionTestCase(unittest.TestCase):
    """Set up the test case for hourly means conversion function"""
    def setUp(self):
        """Create a dataframe containing mock input"""
        self.data = pd.DataFrame(
            index=[0, 1], columns=[
                'date', 'component', 'base', 'hourly_mean_temp'])
        self.data.component = ['I', 'X']
        self.data.base = [53, 200]
        self.data.hourly_mean_temp = [1200, 530]

    def test_hourly_mean_conversion(self):
        """Verify correct conversion to hourly means from mock data"""
        df = io.hourly_mean_conversion(self.data)

        self.assertAlmostEqual(df.iloc[0].hourly_mean, 55)
        self.assertAlmostEqual(df.iloc[1].hourly_mean, 20530)


class AnglesToGeographicTestCase(unittest.TestCase):
    """Set up the test case for conversion from D and H to X and Y"""
    def setUp(self):
        """Create a dataframe containing mock input"""
        self.data = pd.DataFrame(
            index=[0, 1], columns=[
                'date', 'component', 'daily_mean'])
        self.data.component = ['H', 'D']
        self.data.daily_mean = [20530, 55]
        self.data.date = [dt.date(day=15, month=1, year=1963), dt.date(day=15,
                          month=1, year=1963)]
        self.data = self.data.pivot(index='date', columns='component',
                                    values='daily_mean')

    def test_angles_to_geographic(self):
        """Verify correct conversion from mock data"""
        df = io.angles_to_geographic(self.data)

        self.assertAlmostEqual(df.iloc[0].X, 11775.524238286978)
        self.assertAlmostEqual(df.iloc[0].Y, 16817.191469253001)


class WDCXYZTestCase(unittest.TestCase):
    """Set up the test case for conversion to X, Y and Z components"""
    def setUp(self):
        """Create a dataframe containing mock input"""
        self.data = pd.DataFrame(
            index=[0, 1, 2, 3, 4, 5], columns=[
                'date', 'component', 'base', 'hourly_mean_temp'])
        self.data.component = ['H', 'D', 'X', 'Y', 'Z', 'X']
        self.data.base = [200, 53, np.nan, np.nan, 300, 9999]
        self.data.hourly_mean_temp = [530, 1200, np.nan, np.nan, 430, 9999]
        self.data.date.iloc[0:5] = dt.date(day=15, month=1, year=1963)
        self.data.date.iloc[5] = dt.date(day=20, month=1, year=1963)

    def test_wdc_xyz(self):
        """Verify correct conversion to X, Y and Z components from mock data"""
        df = io.wdc_xyz(self.data)

        self.assertAlmostEqual(df.iloc[0].X, 11775.524238286978)
        self.assertAlmostEqual(df.iloc[0].Y, 16817.191469253001)
        self.assertAlmostEqual(df.iloc[0].Z, 30430.000000000000)
        self.assertTrue(np.isnan(df.iloc[1].X))

    def test_wdc_xyz_is_nan_if_Z_missing(self):
        """Ensure missing Z data are set to NaN"""
        self.data = self.data[self.data.component != 'Z']
        df = io.wdc_xyz(self.data)
        self.assertTrue(np.isnan(df.iloc[1].Z))

    def test_wdc_xyz_is_nan_if_DHXY_missing(self):
        """Ensure missing X, Y, D or H data are set to NaN"""
        self.data = self.data[~(self.data.component.isin(['D', 'H', 'X',
                                                          'Y']))]
        df = io.wdc_xyz(self.data)
        self.assertTrue(np.isnan(df.iloc[0].X))
        self.assertTrue(np.isnan(df.iloc[0].Y))


class WDCReadTestCase(unittest.TestCase):
    """Set up the test case for reading WDC files"""
    def setUp(self):
        """Create a dataframe containing mock input and specify test file"""
        self.data = pd.DataFrame(columns=['date', 'X', 'Y', 'Z'])
        self.data.columns.name = 'component'
        self.data.date = pd.date_range('1883-01-01 00:30:00', freq='H',
                                       periods=5)
        self.data.X = [np.nan, 18656.736886, 18657.537749, 18660.729592,
                       18658.976990]
        self.data.Y = [np.nan, -5487.438180, -5491.801722, -5480.946278,
                       -5493.994785]
        self.data.Z = [np.nan, np.nan, np.nan, np.nan, np.nan]

        self.filename = os.path.join(TEST_DATA_PATH, 'testdata3.wdc')

    def test_wdc_readfile(self):
        """Verify correct reading of test file by comparing with mock data"""
        df = io.wdc_readfile(self.filename)

        assert_frame_equal(df.head(), self.data)


class WDCAppendTestCase(unittest.TestCase):
    """Set up the test case for appending WDC data"""
    def setUp(self):
        """Create a dataframe containing mock input and specify test file"""
        # the type matches when this is a single value range...
        self.value1 = [pd.date_range('1911-1-1 0:30', '1911-1-1 0:30')]
        self.value2 = [45294.0]
        self.dimensions = (1416, 4)
        self.filename = 'testappenddata'

    def test_append_wdc_data(self):
        """Verify correct append of test file by comparing with mock data"""
        df = io.append_wdc_data(obs_name=self.filename, path=TEST_DATA_PATH)

        self.assertEqual(self.dimensions, df.shape)
        self.assertEqual(self.value1, df['date'].head(1).values)
        self.assertAlmostEqual(self.value2, df['Z'].tail(1).values)


class WDCHourlyToCSVTestCase(unittest.TestCase):
    """
    Mock the calls I don't want to actually make or are tested elsewhere, and
    beware that patches are applied in _reverse_ order, because whoever wrote
    mock is the devil
    """

    def setUp(self):
        """Specify mock filenames and paths"""
        self.wdc_path = 'data'
        self.write_dir = './a-test-path/to-nowhere'
        self.obs_list = ['ESK']
        self.print_obs = True
        self.wdc_data = pd.DataFrame()

    @mock.patch('magpysv.io.write_csv_data')
    @mock.patch('magpysv.io.append_wdc_data')
    @mock.patch('os.makedirs')
    @mock.patch('os.path.exists', return_value=False)
    def test_wdc_to_hourly_csv_path_exists(self, mock_exists, mock_makedirs,
                                           mock_append_wdc_data,
                                           mock_write_csv_data):
        """Check is paths/files exist"""
        io.wdc_to_hourly_csv(wdc_path=self.wdc_path,
                             write_dir=self.write_dir,
                             obs_list=self.obs_list,
                             print_obs=self.print_obs)
        assert mock_makedirs.call_count == 1
        mock_makedirs.assert_called_with(self.write_dir)

    @mock.patch('magpysv.io.write_csv_data')
    @mock.patch('magpysv.io.append_wdc_data')
    @mock.patch('sys.stdout', new_callable=StringIO)
    @mock.patch('os.path.exists', return_value=True)
    def test_wdc_to_hourly_csv_call_print(self, mock_exists, mock_print,
                                          mock_append_wdc_data,
                                          mock_write_csv_data):
        """Check if print function called"""
        io.wdc_to_hourly_csv(wdc_path=self.wdc_path,
                             write_dir=self.write_dir,
                             obs_list=self.obs_list,
                             print_obs=self.print_obs)
        self.assertEqual(self.obs_list[0] + '\n', mock_print.getvalue())

    @mock.patch('magpysv.io.write_csv_data')
    @mock.patch('magpysv.io.append_wdc_data')
    @mock.patch('os.path.exists', return_value=True)
    def test_wdc_to_hourly_csv_call_write(self, mock_exists,
                                          mock_append_wdc_data,
                                          mock_write_csv_data):
        """Check if write function called"""
        mock_append_wdc_data.return_value = self.wdc_data
        io.wdc_to_hourly_csv(wdc_path=self.wdc_path,
                             write_dir=self.write_dir,
                             obs_list=self.obs_list,
                             print_obs=self.print_obs)
        mock_write_csv_data.assert_called_with(data=self.wdc_data,
                                               write_dir=self.write_dir,
                                               obs_name=self.obs_list[0])



# -*- coding: utf-8 -*-
#    Copyright (C) 2016  Grace Cox (University of Liverpool)
#
#    Released under the MIT license, a copy of which is located at the root of
#    this project.
"""Module containing functions to run the COV-OBS (Gillet et al) field model.

Part of the MagPySV package for geomagnetic data analysis. Contains a function
to obtain a complete list of geomagnetic observatory locations from the WDC
webserver and another function to run the COV-OBS magnetic field model by
Gillet et al. (2013, Geochem. Geophys. Geosyst.,
https://doi.org/10.1002/ggge.20041; 2015, Earth, Planets and Space,
https://doi.org/10.1186/s40623-015-0225-z2013) to obtain model
predictions for these observatory locations. The code can be obtained from
http://www.spacecenter.dk/files/magnetic-models/COV-OBSx1/ and no modifications
are necessary to run it using MagPySV.
"""

import os
from subprocess import Popen, PIPE
import numpy as np
import requests


def get_observatory_list():
    """Obtain the complete list of observatory locations held by the WDC.

    Obtains a dictionary containing information for all geomagnetic
    observatories known to the WDC using the BGS website at:
    http://app.geomag.bgs.ac.uk/wdc/

    The following information is given for each location:

    'AAA':
         {
         'code': 'AAA',

         'country': 'Kazakhstan',

         'dataAvailability': {'hour': {'earliest': 1963, 'latest': 2015},

         'minute': {'earliest': 2005, 'latest': 2015}},

         'dateClosed': None,

         'dateOpened': [1963, 1, 1],

         'elevation': 1300.0,

         'latitude': 43.18,

         'longitude': 76.92,

         'name': 'Alma Ata'
         }

    Returns:
        stations (dict):
            dictionary containing information about each geomagnetic
            observatory.
    """

    url_base = r'http://app.geomag.bgs.ac.uk/wdc/'
    station_resource = 'stations'
    stations_url = url_base + station_resource

    response = requests.get(stations_url)
    # Dictionary with IAGA codes as primary keys
    stations = response.json()
    return stations


def run_covobs(*, stations, model_path, output_path):
    """Use observatory latitude, longitude and elevation to run COV-OBS.

    Uses the dictionary of observatory information obtained from the WDC site
    to run the COV-OBS field model Gillet et al. (2013, Geochem. Geophys.
    Geosyst.,
    https://doi.org/10.1002/ggge.20041; 2015, Earth, Planets and Space,
    https://doi.org/10.1186/s40623-015-0225-z2013) for a given location given
    in geodetic coordinates (model output is also in geodetic coordinates).
    Converts latitude in degrees to colatitude in radians, longitude in degrees
    (0 to 360) into radians (-pi to pi) and elevation in m to km. It then runs
    the fortran exectuable for the field model and passes the location data as
    command line arguments. The output files are stored as mf_obs.dat and
    sv_obs.dat for magnetic field and secular variation predictions
    respectively (e.g. mf_ngk.dat and sv_ngk.dat for Niemegk.)

    Assumes that the user has compiled the fortran source code and called the
    executable "a.out".  No modification to the fortran source code is
    required (code can be downloaded from
    http://www.spacecenter.dk/files/magnetic-models/COV-OBSx1/).

    The COV-OBS code can also be used to run other field models if modified to
    accept a different spline file as input, rather than the supplied
    COV-OBS.x1-int file.

    Args:
        stations (dict): dictionary containing information about each
                geomagnetic observatory.
        model_path (str): path to the compiled COV-OBS executable.
        output_path (str): path to the directory in which the model output
            should be stored.
    """
    mycwd = os.getcwd()
    os.chdir(model_path)
    # Create the output directory if it does not exist
    if not os.path.exists(output_path):
        os.makedirs(output_path)

    for ob in stations.keys():
        print(ob)
        # Convert from latitude in degrees to colatitude in radians
        colatitude = np.deg2rad(90.0 - stations.get(ob).get('latitude'))
        # Convert from longitude in degrees (0 to 360) to radians (-pi to pi)
        longitude = np.deg2rad(stations.get(ob).get('longitude') - 360.0)
        # Convert elevation from m to km
        if stations.get(ob).get('elevation') is None:
            altitude = 0.0
        else:
            altitude = stations.get(ob).get('elevation')/1000.0
        # Create a string containing the inputs to the COV-OBS field model
        model_inputs = "%s\n%s\n%s\n" % (str(altitude), str(colatitude),
                                         str(longitude))
        # Create a process so python can interact with the model executable
        p = Popen('./a.out', stdin=PIPE,
                  stdout=PIPE)
        # Pass the altitude, colatitude and longitude to the field model
        p.communicate(model_inputs.encode())
        p.wait()
        # Rename the output files so they contain the observatory name
        os.rename('mfpred.dat', os.path.join(output_path,
                  'mf_%s.dat' % ob.lower()))
        os.rename('svpred.dat', os.path.join(output_path,
                  'sv_%s.dat' % ob.lower()))
    # Return to previous working directory
    os.chdir(mycwd)



# -*- coding: utf-8 -*-
#    Copyright (C) 2016  Grace Cox (University of Liverpool)
#
#    Released under the MIT license, a copy of which is located at the root of
#    this project.
"""Module containing functions to parse files output by magnetic field models.

Part of the MagPySV package for geomagnetic data analysis. This module provides
various functions to read SV files output by geomagnetic field models.
"""


import pandas as pd
import numpy as np
import os


def calculate_sv(obs_data, mean_spacing=1):
    """Calculate the secular variation from the observed magnetic field values.

    Uses monthly means of geomagnetic observatory data to calculate the SV
    according to user-specified sampling. The typical choices are monthly
    differences of monthly means and annual differences of monthly means. For
    samplings other than monthly differences, the datetime objects of the
    calculated SV are taken the midpoint of the datetime objects of the field
    data. E.g. differencing the means of the field January 1999 and in January
    2000 yields the SV for July 1999.

    Args:
        obs_data (pandas.DataFrame): dataframe containing means (usually
            monthly) of observed geomagnetic field values.
        mean_spacing (int): the number of months separating the monthly mean
            values used to calculate the SV. Set to 1 (default) to use adjacent
            months of data (first differences of monthly means) or set to 12
            to calculate annual differences of monthly means.

    Returns:
        obs_sv (pandas.DataFrame): dataframe containing SV data.
    """
    obs_sv = pd.DataFrame()
    # Set the SV month to halfway between the MF measurement dates
    obs_sv['date'] = obs_data['date'] - pd.tseries.offsets.DateOffset(
        months=int(np.floor(mean_spacing/2)), day=1)
    if mean_spacing % 2 == 1:
        obs_sv['date'] = obs_sv['date'].apply(lambda dte: dte.replace(day=1))
    else:
        obs_sv['date'] = obs_sv['date'].apply(lambda dte: dte.replace(day=15))

    # Calculate scale required to give SV in nT/yr
    scaling_factor = 12 / mean_spacing
    obs_sv['dx'] = scaling_factor * obs_data['X'].diff(periods=mean_spacing)
    obs_sv['dy'] = scaling_factor * obs_data['Y'].diff(periods=mean_spacing)
    obs_sv['dz'] = scaling_factor * obs_data['Z'].diff(periods=mean_spacing)
    obs_sv.drop(obs_sv.head(mean_spacing).index, inplace=True)

    return obs_sv


def calculate_residuals(*, obs_data, model_data):
    """Calculate SV residuals (observed - prediction) using datetime objects.

    Args:
        obs_data (pandas.DataFrame): dataframe containing means (usually
            monthly) of SV calculated from observed geomagnetic field values.
        model_data (pandas.DataFrame): dataframe containing the SV predicted by
            a geomagnetic field model.

    Returns:
        residuals (pandas.DataFrame): dataframe containing SV residuals.
    """
    dates = obs_data['date']
    model_data = model_data[model_data['date'].isin(dates)]
    obs_data.drop(obs_data.columns[[0]], axis=1, inplace=True)
    model_data.drop(model_data.columns[[0]], axis=1, inplace=True)

    # Calculate SV residuals (data - model prediction) for all observatories
    residuals = pd.DataFrame(
        obs_data.values - model_data.values,
        columns=obs_data.columns)
    obs_data.insert(0, 'date', dates)

    return residuals


def data_resampling(data, sampling='MS', average_date=True):
    """Resample the hourly geomagnetic data to a specified frequency.

    Args:
        data (pandas.DataFrame): dataframe containing datetime objects and
            hourly means of magnetic data.
        sampling (str): new sampling frequency. Default value is 'MS'
            (monthly means), which averages data for each month and sets
            the datetime object to the first day of that month. Use 'M'
            to set the datetime object to the final day of the month.
            Another useful option is 'A' (annual means), which averages
            data for a whole year and sets the datetime object to the final
            day of the year. Use 'AS' to set the datetime object to the
            first day of the year.
        average_date (bool): the specified resampling intervals only have
            options for setting the date to the first ('MS' and 'AS') or final
            ('M' and 'A') day of the month or year. For monthly averages, a
            more appropriate representative date is the middle of that month
            (i.e. the 15th day of the month). For annual averages, an
            appropriate representative date is the middle of that year (taken
            as July 1st of the year.) This option is used to set the dates to
            the centre of the sampling interval. Defaults to True.

    Returns:
        data (pandas.DataFrame):
            dataframe of datetime objects and monthly/annual means of
            observatory data.
    """
    resampled = data.set_index('date', drop=False).resample(sampling).mean()
    resampled.reset_index(inplace=True)
    if average_date is True and 'A' in sampling:
        resampled['date'] = resampled['date'] + pd.tseries.offsets.DateOffset(
                            month=7, day=1)
    if average_date is True and 'M' in sampling:
        resampled['date'] = resampled['date'] + pd.tseries.offsets.DateOffset(
                            day=15)
    return resampled


def apply_Ap_threshold(*, Ap_file=None, obs_data, threshold):
    """Remove observatory data for times with ap values above threshold value.

    Args:
        Ap_file (str): path to file containing hourly values for the ap index.
        obs_data (pandas.DataFrame): dataframe containing hourly means of
            observed geomagnetic field values.
        threshold (int): the threshold Ap value. Data for days with a higher Ap
            value will be replaced with NaNs and omitted from monthly (or
            annual etc) means.

    Returns:
        obs_data (pandas.DataFrame): data with ap threshold applied.
    """
    col_names = ['date', 'Ap']
    Ap_hourly = pd.read_csv(Ap_file, names=col_names,
                            dtype={'date': 'str', 'Ap': 'float'},
                            parse_dates=[0], skiprows=0, index_col=None)
    # Merge the two dataframes so that only dates contained within both are
    # retained
    obs_data = pd.merge(obs_data, Ap_hourly, on='date', how='inner')
    # Use the threshold to discard data from days with high external field
    # activity
    obs_data.loc[obs_data.Ap > threshold, 'X'] = np.NaN
    obs_data.loc[obs_data.Ap > threshold, 'Y'] = np.NaN
    obs_data.loc[obs_data.Ap > threshold, 'Z'] = np.NaN
    obs_data.drop(['Ap'], axis=1, inplace=True)

    return obs_data


def remove_selected_points(*, data, fname):
    """Remove specified points from dataset based on list of points in a file.

    Reads a list of unwanted points from a file and removes them from the
    dataframe if present. E.g. If the user had monthly SV means and wished to
    exclude the X value at NGK in January 2015 from the analysis, the following
    line would be written in a file

    2015-01-01,ngk,X

    It is preferable (and more repeatable) to use the included outlier
    detection algorithms to remove outliers.

    Args:
        data (pandas.DataFrame): dataframe containing datetime objects and
              daily means of magnetic data.
        fname (str): path to file containing a list of unwanted data points.

    Returns:
        data (pandas.DataFrame): the same dataframe with the points removed.
    """
    col_names = ['date', 'observatory', 'component']
    points = pd.read_csv(fname, sep=',', header=0, names=col_names,
                         parse_dates=[0])

    for data_point in points.itertuples():
        col_name = data_point.component + '_' + data_point.observatory
        try:
            data.loc[data['date'] == data_point.date, col_name] = np.nan
        except KeyError:
            pass

    return data


def calculate_sv_index(obs_data, mean_spacing=1):
    """Calculate the secular variation of a geomagnetic index.

    Uses monthly means of a geomagnetic index (e.g. Dst) to calculate the SV
    according to user-specified sampling. The typical choices are monthly
    differences of monthly means and annual differences of monthly means. For
    samplings other than monthly differences, the datetime objects of the
    calculated SV are taken the midpoint of the datetime objects of the field
    data. E.g. differencing the means of the field January 1999 and in January
    2000 yields the SV for July 1999.

    Args:
        obs_data (pandas.DataFrame): dataframe containing means (usually
            monthly) of observed geomagnetic field values.
        mean_spacing (int): the number of months separating the monthly mean
            values used to calculate the SV. Set to 1 (default) to use adjacent
            months of data (first differences of monthly means) or set to 12 to
            calculate annual differences of monthly means.

    Returns:
        obs_sv (pandas.DataFrame): dataframe containing SV data.
    """
    obs_sv = pd.DataFrame()
    # Set the SV month to halfway between the MF measurement dates
    obs_sv['date'] = obs_data['date'] - pd.tseries.offsets.DateOffset(
        months=int(np.floor(mean_spacing/2)), day=1)
    if mean_spacing % 2 == 1:
        obs_sv['date'] = obs_sv['date'].apply(lambda x: x.replace(day=1))
    else:
        obs_sv['date'] = obs_sv['date'].apply(lambda x: x.replace(day=15))
    # Calculate scale required to give SV in nT/yr
    scaling_factor = 12 / mean_spacing
    obs_sv['mean'] = scaling_factor * obs_data['mean'].diff(
            periods=mean_spacing)
    obs_sv.drop(obs_sv.head(mean_spacing).index, inplace=True)

    return obs_sv


def calculate_correlation_index(*, dates, signal, index_file):
    """Calculate correlation coefficient between signal and a geomagnetic index

    Args:
        dates (datetime.datetime): dates of time series measurements.
        signal (time series): data to be compared with geomagnetic index.
        index_file (str): path to file containing geomagnetic index.

    Returns:
        (tuple): tuple containing:

        - coeff (*float*):
            correlation coefficient.
        - merged (*pandas.DataFrame*):
            dataframe containing dates, signal and the geomagnetic index.
    """
    col_names = ['date', 'index_vals']
    index_df = pd.read_csv(index_file, names=col_names,
                           dtype={'date': 'str', 'index_vals': 'float'},
                           parse_dates=[0], skiprows=1, index_col=None)
    df = pd.DataFrame({'date': dates, 'proxy': signal})
    # Merge the two dataframes so that only dates contained within both are
    # retained
    merged = pd.merge(df.dropna(), index_df.dropna(), on='date', how='inner')
    coeff = np.corrcoef(merged.index_vals, merged.proxy)

    return np.abs(coeff.data[0, 1]), merged


def get_baseline_info(*, fname=None):
    """Read documented baseline changes from a file.
    Args:
        fname (str): location of file containing documented baseline changes.

    Returns:
        data (pandas.DataFrame): baseline change data.
    """
    if fname is None:
        fname = os.path.join(os.path.dirname(__file__), 'baseline_records')

    col_names = ['observatory', 'jump_year', 'x_jump', 'y_jump', 'z_jump']
    data = pd.read_csv(fname, sep=',', names=col_names)
    data['jump_year'] = pd.to_datetime(dict(year=data['jump_year'],
                                       month=1, day=1))
    return data


def correct_baseline_change(*, observatory, field_data, baseline_data,
                            print_data):
    """Correct documented baseline changes.

    Args:
        observatory (str): observatory name given as a 3-digit IAGA code.
        field_data (pandas.DataFrame): uncorrected magnetic field data.
        baseline_data (pandas.DataFrame): baseline discontinuity data in the
            format output by get_baseline_info.
        print_data (bool): option to print the corrections made.
    """
    # Extract baseline data for the specified observatory
    obs_jumps = baseline_data.loc[
        baseline_data['observatory'] == str.upper(observatory)]

    if print_data is True:
        print(obs_jumps)
    # Loop over a baseline changes and make the correction
    for jump in obs_jumps.index:
        # Check if a change of zero is recorded
        if (obs_jumps.loc[jump].x_jump == 0 & obs_jumps.loc[jump].y_jump == 0 &
                obs_jumps.loc[jump].z_jump == 0):
            print("Baseline change of unknown magnitude: ",
                  obs_jumps.loc[jump].jump_year)
        else:
            field_data.loc[field_data['date'] < obs_jumps.loc[jump].jump_year,
                           'X'] = field_data.loc[
                field_data['date'] < obs_jumps.loc[jump].jump_year, 'X'] - \
                obs_jumps.loc[jump].x_jump
            field_data.loc[field_data['date'] < obs_jumps.loc[jump].jump_year,
                           'Y'] = field_data.loc[
                field_data['date'] < obs_jumps.loc[jump].jump_year, 'Y'] - \
                obs_jumps.loc[jump].y_jump
            field_data.loc[field_data['date'] < obs_jumps.loc[jump].jump_year,
                           'Z'] = field_data.loc[
                field_data['date'] < obs_jumps.loc[jump].jump_year, 'Z'] - \
                obs_jumps.loc[jump].z_jump



# -*- coding: utf-8 -*-
"""
Created on Thu Feb  2 16:45:42 2017

Testing functions for tools.py.

@author: Grace Cox and Will Brown
"""
import unittest
import os
from .. import tools
import pandas as pd
import datetime as dt


class DataResamplingTestCase(unittest.TestCase):
    """Set up test case for data resampling"""
    def setUp(self):
        """Specify location of test file"""
        # Directory where the test files are located
        self.path = os.path.join(os.path.dirname(__file__), 'data')
        testfile = os.path.join(self.path, 'testdaily.csv')
        self.col_names = ['date', 'code', 'component', 'daily_mean']
        self.data = pd.read_csv(testfile, sep=' ', header=0,
                                names=self.col_names, parse_dates=[0])
        self.averaged = tools.data_resampling(self.data)

    def test_data_resampling(self):
        """Verify correct resampling of test file data"""
        self.assertAlmostEqual(self.averaged.daily_mean.values[0], 801.000000)
        self.assertAlmostEqual(self.averaged.daily_mean.values[7],
                               33335.750000)
        self.assertAlmostEqual(self.averaged.daily_mean.values[-1],
                               45115.500000)
        self.assertEqual(self.averaged.date[0], dt.datetime(day=15, month=1,
                         year=2000))
        self.assertEqual(self.averaged.date[1], dt.datetime(day=15, month=2,
                         year=2000))
        self.assertEqual(self.averaged.date[7], dt.datetime(day=15, month=8,
                         year=2000))



# -*- coding: utf-8 -*-
#    Copyright (C) 2016  Grace Cox (University of Liverpool)
#
#    Released under the MIT license, a copy of which is located at the root of
#    this project.
"""Module containing functions to remove external signal from geomagnetic data.

Part of the MagPySV package for geomagnetic data analysis. This module provides
various functions to denoise geomagnetic data by performing principal component
analysis and identifying and removing outliers. Also contains an outlier
detection function based on median absolute deviation from the median (MAD).
"""


import pandas as pd
import magpysv.plots as plots
import numpy as np
from sklearn.decomposition import PCA as sklearnPCA
from sklearn.preprocessing import Imputer


def eigenvalue_analysis_impute(*, dates, obs_data, model_data, residuals,
                               proxy_number=1):
    """Remove external signal from SV data using Principal Component Analysis.

    Perform principal component analysis (PCA) on secular variation
    residuals (the difference between the observed SV and that predicted by a
    geomagnetic field model) calculated from annual differences of monthly
    means at several observatories. Uses the imputer from sklearn.preprocessing
    to fill in missing data points and calculates the singular values of the
    data matrix for n observatories (uses Singular Values Decomposition, SVD).
    The residuals are rotated into the eigendirections and denoised using the
    method detailed in Wardinski & Holme (2011, GJI,
    https://doi.org/10.1111/j.1365-246X.2011.04988.x). The SV residuals of the
    noisy component for all observatories combined are used as a proxy for the
    unmodelled external signal. The denoised data are then rotated back into
    geographic coordinates. The pca algorithm outputs the singular values
    (these are equal to the square root of the eigenvalues of the covariance
    matrix) sorted from largest to smallest, so the corresponding eigenvector
    matrix has the 'noisy' direction in the first column and the 'clean'
    direction in the final column.

    Note that the SVD algorithm cannot be used if any data are missing, which
    is why imputation is needed with this method. The function
    denoise.eigenvalue_analysis permits missing values and does not
    infill them - that is the more robust function.

    Smallest eigenvalue: 'quiet' direction

    Largest eiegenvalue: 'noisy' direction

    Args:
        dates (datetime.datetime): dates of the time series measurements.
        obs_data (pandas.DataFrame): dataframe containing columns for
            monthly/annual means of the X, Y and Z components of the secular
            variation at the observatories of interest.
        model_data (pandas.DataFrame): dataframe containing columns for field
            model prediction of the X, Y and Z components of the secular
            variation at the same observatories as in obs_data.
        residuals (pandas.DataFrame): dataframe containing the SV residuals
            (difference between the observed data and model prediction).
        proxy_number (int): the number of 'noisy' directions used to create
            the proxy for the external signal removal. Default value is 1 (only
            the residual in the direction of the largest eigenvalue is used).
            Using n directions means that proxy is the sum of the SV residuals
            in the n noisiest eigendirections.

    Returns:
        (tuple): tuple containing:

        - denoised_sv (*pandas.DataFrame*):
            dataframe with dates in the first
            column and columns for the denoised X, Y and Z secular variation
            components at each of the observatories for which data were
            provided.
        - proxy (*array*):
            the signal that was used as a proxy for unmodelled
            external magnetic field in the denoising stage.
        - eig_values (*array*):
            the singular values of the obs_data matrix.
        - eig_vectors (*array*):
            the eigenvectors associated with the n largest
            singular values of the data matrix. For example, if the residuals
            in the two 'noisiest' directions are used as the proxy for external
            signal, then these two eigenvectors are returned.
        - projected_residuals (*array*):
            SV residuals rotated into the eigendirections.
        - corrected_residuals (*array*):
            SV residuals after the denoising process.
    """
    # Fill in missing SV values (indicated as NaN in the data files)
    imp = Imputer(missing_values='NaN', strategy='mean', axis=0)
    imputed_residuals = imp.fit_transform(residuals)

    pca = sklearnPCA()
    projected_residuals = pca.fit_transform(imputed_residuals)
    eig_values = pca.explained_variance_
    eig_vectors = pca.components_

    # Use the method of Wardinski & Holme (2011) to remove unmodelled external
    # signal in the SV residuals. The variable 'proxy' contains the noisy
    # component residual for all observatories combined

    corrected_residuals = []

    if proxy_number == 1:
        noisy_direction = eig_vectors[:, 0]
        proxy = projected_residuals[:, 0]
        for idx in range(len(proxy)):
            corrected_residuals.append(
                imputed_residuals.data[idx, :] - proxy[idx] * noisy_direction)
    elif proxy_number > 1:
        noisy_direction = eig_vectors[:, 0:proxy_number]
        proxy = np.sum(projected_residuals[:, 0:proxy_number], axis=1)
        for idx in range(len(projected_residuals[:, 0])):
            corrected = imputed_residuals.data[idx, :]
            for direction in range(proxy_number):
                corrected = corrected - projected_residuals[idx, direction] \
                    * noisy_direction[:, direction]
            corrected_residuals.append(corrected)

    corrected_residuals = pd.DataFrame(corrected_residuals,
                                       columns=obs_data.columns)
    denoised_sv = pd.DataFrame(
        corrected_residuals.values + model_data.values,
        columns=obs_data.columns)
    denoised_sv.insert(0, 'date', dates)

    return denoised_sv, proxy, eig_values, eig_vectors, projected_residuals,\
        corrected_residuals.astype('float')


def eigenvalue_analysis(*, dates, obs_data, model_data, residuals,
                        proxy_number=1):
    """Remove external signal from SV data using principal Component Analysis.

    Perform principal component analysis (PCA) on secular variation
    residuals (the difference between the observed SV and that predicted by a
    geomagnetic field model) calculated from annual differences of monthly
    means at several observatories. Uses masked arrays to discount missing data
    points and calculates the eigenvalues/vectors of the (3nx3n) covariance
    matrix for n observatories. The residuals are rotated into the
    eigendirections and denoised using the method detailed in Wardinski & Holme
    (2011, GJI, https://doi.org/10.1111/j.1365-246X.2011.04988.x). The SV
    residuals of the noisy component for all observatories
    combined are used as a proxy for the unmodelled external signal. The
    denoised data are then rotated back into geographic coordinates. The PCA
    algorithm outputs the eigenvalues sorted from largest to smallest, so the
    corresponding eigenvector matrix has the 'noisy' direction in the first
    column and the 'clean' direction in the final column.

    This algorithm masks missing data so that they are not taken into account
    during the PCA. Missing values are not infilled or estimated, so NaN
    values in the input dataframe are given as NaN values in the output.

    Smallest eigenvalue 'quiet' direction

    Largest eiegenvalue 'noisy' direction

    Args:
        dates (datetime.datetime): dates of the time series measurements.
        obs_data (pandas.DataFrame): dataframe containing columns for
            monthly/annual means of the X, Y and Z components of the secular
            variation at the observatories of interest.
        model_data (pandas.DataFrame): dataframe containing columns for field
            model prediction of the X, Y and Z components of the secular
            variation at the same observatories as in obs_data.
        residuals (pandas.DataFrame): dataframe containing the SV residuals
            (difference between the observed data and model prediction).
        proxy_number (int): the number of 'noisy' directions used to create
            the proxy for the external signal removal. Default value is 1 (only
            the residual in the direction of the largest eigenvalue is used).
            Using n directions means that proxy is the sum of the SV residuals
            in the n noisiest eigendirections.

    Returns:
        (tuple): tuple containing:

        - denoised_sv (*pandas.DataFrame*):
            dataframe with datetime objects in the
            first column and columns for the denoised X, Y and Z SV components
            at each of the observatories for which data were provided.
        - proxy (*array*):
            the signal that was used as a proxy for unmodelled
            external magnetic field in the denoising stage.
        - eig_values (*array*):
            the eigenvalues of the obs_data matrix.
        - eig_vectors (*array*):
            the eigenvectors associated with the n largest
            eigenvalues of the data matrix. For example, if the residuals
            in the two 'noisiest' directions are used as the proxy for external
            signal, then these two eigenvectors are returned.
        - projected_residuals (*array*):
            SV residuals rotated into the eigendirections.
        - corrected_residuals (*array*):
            SV residuals after the denoising process.
    """
    # Create a masked version of the residuals array so that we can perform the
    # PCA ignoring all nan values
    masked_residuals = np.ma.array(residuals, mask=np.isnan(residuals))

    # Calculate the covariance matrix of the masked residuals array
    covariance_matrix = np.ma.cov(masked_residuals, rowvar=False,
                                  allow_masked=True)
    # Calculate the eigenvalues and eigenvectors of the covariance matrix
    eig_values, eig_vectors = np.linalg.eig(covariance_matrix)
    # Sort the eigenvalues in decreasing order
    idx = np.argsort(eig_values)[::-1]
    eig_values = eig_values[idx]
    # Sort the eigenvectors according to the same index
    eig_vectors = eig_vectors[:, idx]

    # Project the residuals onto the eigenvectors
    projected_residuals = np.ma.dot(masked_residuals, eig_vectors)

    # Use the method of Wardinski & Holme (2011) to remove unmodelled external
    # signal in the SV residuals. The variable 'proxy' contains the noisy
    # component residual for all observatories combined

    corrected_residuals = []

    if proxy_number == 1:
        noisy_direction = eig_vectors[:, 0]
        proxy = projected_residuals[:, 0]
        for idx in range(len(proxy)):
            corrected_residuals.append(
                masked_residuals.data[idx, :] - proxy[idx] * noisy_direction)
    elif proxy_number > 1:
        noisy_direction = eig_vectors[:, 0:proxy_number]
        proxy = np.sum(projected_residuals[:, 0:proxy_number], axis=1)
        for idx in range(len(projected_residuals[:, 0])):
            corrected = masked_residuals.data[idx, :]
            for direction in range(proxy_number):
                corrected = corrected - projected_residuals[idx, direction] \
                    * noisy_direction[:, direction]
            corrected_residuals.append(corrected)

    corrected_residuals = pd.DataFrame(corrected_residuals,
                                       columns=obs_data.columns)
    denoised_sv = pd.DataFrame(
        corrected_residuals.values + model_data.values,
        columns=obs_data.columns)

    denoised_sv.insert(0, 'date', dates)

    return denoised_sv, proxy, eig_values, eig_vectors, projected_residuals,\
        corrected_residuals.astype('float')


def detect_outliers(*, dates, signal, obs_name, window_length, threshold,
                    signal_type='SV', plot_fig=False, save_fig=False,
                    write_path=None, fig_size=(8, 6), font_size=12,
                    label_size=16):
    """Detect outliers in a time series and remove them.

    Use the median absolute deviation from the median (MAD) to identify
    outliers. The time series are long and highly variable so it is not
    appropriate to use single values of median to represent the whole series.
    The function uses a running median to better characterise the series
    (the window length and a threshold value stating many MADs from the median
    a point must be before it is classed as an outlier are user-specified).

    Args:
        dates (datetime.datetime): dates of the time series measurements.
        signal (array): array (or column from a pandas.DataFrame) containing
            the time series of interest.
        obs_name (str): states the component of interest and the three digit
           IAGA observatory name.
        window_length (int): number of months over which to take the running
            median.
        threshold (float): the minimum number of median absolute deviations a
            point must be away from the median in order to be considered an
            outlier.
        signal_type (str): specify whether magnetic field ('MF') or secular
            variation ('SV') is plotted. Defaults to SV.
        plot_fig (bool): option to plot figure of the time series and
            identified outliers. Defaults to False.
        save_fig (bool): option to save figure if plotted. Defaults to False.
        write_path (str): output path for figure if saved.
        fig_size (array): figure size in inches. Defaults to 8 inches by 6
            inches.
        font_size (int): font size for axes. Defaults to 12 pt.
        label_size (int): font size for axis labels. Defaults to 16 pt.

    Returns:
        signal (array):
            the input signal with identified outliers removed (set to NaN).
    """
    signal_temp = pd.DataFrame(data=signal.copy())
    # Account for missing values when using rolling_median and rolling_std.
    # ffill (bfill) propagates the closest value forwards (backwards) through
    # nan values. E.g. [np.nan, np.nan, 1, 9, 7, np.nan, np.nan] returns as
    # [1, 1, 1, 9, 7, 7, 7]. The limit of half the window length is used so the
    # first ffill cannot overwrite the beginning of the next valid interval
    # (bfill values are used there instead).
    signal_temp = signal_temp.ffill(limit=int(window_length / 2 + 1)).bfill()
    # calculate the running median and median absolute standard deviation
    running_median = signal_temp.rolling(window=window_length,
                                         center=True).median().bfill().ffill()
    diff = (signal_temp - running_median).abs()
    med_abs_deviation = diff.rolling(window=window_length,
                                     center=True).median().bfill().ffill()
    # Normalise the median abolute deviation
    modified_z_score = diff / med_abs_deviation
    # Identify outliers
    outliers = signal_temp[modified_z_score > threshold]
    # Plot the outliers and original time series if required
    if plot_fig is True:
        plots.plot_outliers(dates=dates, obs_name=obs_name, signal=signal,
                            outliers=outliers, save_fig=save_fig,
                            write_path=write_path, fig_size=fig_size,
                            font_size=font_size, label_size=label_size,
                            signal_type=signal_type)
    # Set the outliers to NaN
    idx = np.where(modified_z_score > threshold)[0]
    signal.iloc[idx] = np.nan
    return signal.astype('float')



"""minimal package setup"""
import os
from setuptools import setup, find_packages

# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name="magpysv",
    version="1.0",
    author="Grace Cox",
    author_email="grace.cox@dias.ie",
    license="MIT",
    url="https://github.com/gracecox/MagPySV",
    description="Download, process and denoise geomagnetic observatory data",
    long_description=read('readme.rst'),
    packages=find_packages(),
    include_package_data=True,
    package_data={"magpysv": ["baseline_records"]},
    classifiers=["Programming Language :: Python :: 3"],
    zip_safe=False,
    install_requires=['jupyter>=1.0.0','matplotlib>=2.0.0','notebook>=4.3.1',
    'numpy>=1.12.0','pandas>=0.19.2','requests>=2.12.4','scikit-learn>=0.18.1',
    'scipy>=0.18.1','gmdata_webinterface'],
    extras_require={'develop': ['jupyter>=1.0.0','matplotlib>=2.0.0','notebook>=4.3.1',
    'numpy>=1.12.0','pandas>=0.19.2','requests>=2.12.4','scikit-learn>=0.18.1',
    'scipy>=0.18.1','prospector>=0.12.7','pytest>=3.0.6',
    'pytest-cov>=2.4.0','Sphinx>=1.5.1','sphinx-rtd-theme>=0.1.9','gmdata_webinterface']}
)
